# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions for dealing with regions and bounding boxes.
"""

import numpy as np

from ._validation import check_coordinates, check_region


def pad_region(region, pad):
    """
    Extend the borders of a region by the given amount.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.
    pad : float or tuple = (pad_WE, pad_SN, ...)
        The amount of padding to add to the region. If it's a single number,
        add this to all boundaries of region equally. If it's a tuple of
        numbers, then will add different padding to each dimension of the
        region respectively. If a tuple, the number of elements should be half
        of the number of elements in *region*.

    Returns
    -------
    padded_region : tuple = (W, E, S, N, ...)
        The padded region.

    Examples
    --------
    >>> pad_region((0, 1, -5, -3), 1)
    (-1, 2, -6, -2)
    >>> pad_region((0, 1, -5, -3, 6, 7), 1)
    (-1, 2, -6, -2, 5, 8)
    >>> pad_region((0, 1, -5, -3), (2, 3))
    (-2, 3, -8, 0)
    >>> pad_region((0, 1, -5, -3, 6, 7), (2, 3, 1))
    (-2, 3, -8, 0, 5, 8)

    """
    check_region(region)
    ndims = len(region) // 2
    if np.isscalar(pad):
        pad = tuple(pad for _ in range(ndims))
    if len(pad) != ndims:
        message = (
            f"Invalid padding '{pad}'. "
            f"Should have {ndims} elements for region '{region}'."
        )
        raise ValueError(message)
    region_pairs = np.reshape(region, (len(region) // 2, 2))
    padded = [[lower - p, upper + p] for p, (lower, upper) in zip(pad, region_pairs)]
    return tuple(np.ravel(padded).tolist())


def get_region(coordinates):
    """
    Get the bounding region of the given coordinates.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape.

    Returns
    -------
    region : tuple = (W, E, S, N, ...)
        The boundaries that contain the coordinates. The order of lower and
        upper boundaries returned follows the order of *coordinates*.

    Examples
    --------
    >>> get_region(([0, 0.5, 1], [-10, -8, -6]))
    (0.0, 1.0, -10.0, -6.0)
    >>> get_region(([0, 0.5, 1], [-10, -8, -6], [4, 10, 16]))
    (0.0, 1.0, -10.0, -6.0, 4.0, 16.0)

    """
    coordinates = check_coordinates(coordinates)
    region = tuple(np.ravel([[np.min(c), np.max(c)] for c in coordinates]).tolist())
    return region


def inside(coordinates, region):
    """
    Determine which points fall inside a given region.

    Points at the boundaries are counted as being inside the region.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Should be in an
        order compatible with the order of boundaries in *region*. Arrays can
        be Python lists. Arrays can be of any shape but must all have the same
        shape.
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.

    Returns
    -------
    are_inside : array of booleans
        An array of booleans with the same shape as the input coordinate
        arrays. Will be ``True`` if the respective coordinates fall inside the
        area, ``False`` otherwise.

    Examples
    --------
    >>> east = [1, 2, 3, 4, 5, 6]
    >>> north = [10, 11, 12, 13, 14, 15]
    >>> region = (2.5, 5.5, 12, 15)
    >>> print(inside((east, north), region))
    [False False  True  True  True False]
    >>> # This also works for 2D-arrays
    >>> east = [[1, 2, 3],
    ...         [1, 2, 3],
    ...         [1, 2, 3]]
    >>> north = [[5, 5, 5],
    ...          [7, 7, 7],
    ...          [9, 9, 9]]
    >>> region = (0.5, 2.5, 7, 9)
    >>> print(inside((east, north), region))
    [[False False False]
     [ True  True False]
     [ True  True False]]
    >>> # and 3D-arrays or higher dimensions
    >>> east = [[[1, 2, 3],
    ...          [1, 2, 3],
    ...          [1, 2, 3]],
    ...         [[1, 2, 3],
    ...          [1, 2, 3],
    ...          [1, 2, 3]]]
    >>> north = [[[5, 5, 5],
    ...           [7, 7, 7],
    ...           [9, 9, 9]],
    ...          [[5, 5, 5],
    ...           [7, 7, 7],
    ...           [9, 9, 9]]]
    >>> up = [[[4, 4, 4],
    ...        [4, 4, 4],
    ...        [4, 4, 4]],
    ...       [[6, 6, 6],
    ...        [6, 6, 6],
    ...        [6, 6, 6]]]
    >>> region = (0.5, 2.5, 7, 9, 4.5, 7)
    >>> print(inside((east, north, up), region))
    [[[False False False]
      [False False False]
      [False False False]]
    <BLANKLINE>
     [[False False False]
      [ True  True False]
      [ True  True False]]]

    """
    check_region(region)
    coordinates = check_coordinates(coordinates)
    ndims = len(region) // 2
    if len(coordinates) != ndims:
        message = (
            f"Invalid coordinates. Expected {ndims} coordinates for region '{region}' "
            f"but got {len(coordinates)} instead."
        )
        raise ValueError(message)
    region_pairs = np.reshape(region, (ndims, 2))
    shape = coordinates[0].shape
    # Allocate temporary arrays to minimize memory allocation overhead
    tmp = tuple(np.empty(shape, dtype=bool) for i in range(2))
    are_inside = np.full(shape, True)
    for coordinate, (lower, upper) in zip(coordinates, region_pairs):
        # Using the logical functions is a lot faster than & > < for some reason.
        # Plus, this way avoids repeated allocation of intermediate arrays by using
        # the "out" argument.
        in_dimension = np.logical_and(
            np.greater_equal(coordinate, lower, out=tmp[0]),
            np.less_equal(coordinate, upper, out=tmp[1]),
            out=tmp[0],
        )
        are_inside = np.logical_and(are_inside, in_dimension, out=are_inside)
    return are_inside



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions for validating inputs and outputs.
"""

import numpy as np


def check_coordinates(coordinates):
    """
    Check that coordinate arrays all have the same shape.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape.

    Returns
    -------
    coordinates : tuple = (easting, northing, ...)
        Tuple of coordinate arrays, converted to numpy arrays if necessary.

    Raises
    ------
    ValueError
        If the coordinates don't have the same shape.
    """
    coordinates = tuple(np.asarray(c) for c in coordinates)
    shapes = [c.shape for c in coordinates]
    if not all(shape == shapes[0] for shape in shapes):
        message = (
            "Invalid coordinates. All coordinate arrays must have the same shape. "
            f"Given coordinate shapes: {shapes}"
        )
        raise ValueError(message)
    return coordinates


def check_coordinates_geographic(coordinates):
    """
    Check if geographic coordinates are within allowed bounds.

    Longitude boundaries should be in range [0, 360] or [-180, 180] and
    latitude boundaries should be in range [-90, 90].

    Parameters
    ----------
    coordinates : tuple = (longitude, latitude)
        Tuple of arrays with the longitude and latitude coordinates of each
        point. Arrays can be Python lists or any numpy-compatible array type.
        Arrays can be of any shape but must all have the same shape.

    Raises
    ------
    ValueError
        If the coordinates are outside their valid ranges or if there aren't
        exactly two coordinates.

    """
    if len(coordinates) != 2:
        message = (
            "Invalid coordinates. Must have exactly 2 elements"
            f" (longitude, latitude), but {len(coordinates)} were given."
        )
        raise ValueError(message)
    longitude, latitude = (np.atleast_1d(c) for c in coordinates)
    west, east = longitude.min(), longitude.max()
    south, north = latitude.min(), latitude.max()
    if west < -180 or east > 360 or (west < 0 and east > 180):
        message = (
            f"Invalid longitude range [{west}, {east}]. "
            "Longitude range must be [0, 360] or [-180, 180]."
        )
        raise ValueError(message)
    if south < -90 or north > 90:
        message = (
            f"Invalid latitude range [{south}, {north}]. "
            "Latitude range must be [-90, 90]."
        )
        raise ValueError(message)


def check_region(region):
    """
    Check that the given region is valid.

    A region is a bounding box for n-dimensional coordinates. There should be
    an even number of elements and lower boundaries should not be larger than
    upper boundaries.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.

    Raises
    ------
    ValueError
        If the region doesn't have even number of entries and any lower
        boundary is larger than the upper boundary.

    """
    if not region or len(region) % 2 != 0:
        message = (
            f"Invalid region '{region}'. Must have an even number of elements, "
            "a lower and an upper boundary for each dimension."
        )
        raise ValueError(message)
    region_pairs = np.reshape(region, (len(region) // 2, 2))
    offending = [lower > upper for lower, upper in region_pairs]
    if any(offending):
        bad_bounds = []
        for dimension, is_bad in enumerate(offending):
            if is_bad:
                lower, upper = region_pairs[dimension]
                bad_bounds.append(f"{dimension} ({lower} > {upper})")
        message = (
            f"Invalid region '{region}'. Lower boundary larger than upper boundary "
            f"in dimension(s): {'; '.join(bad_bounds)}"
        )
        raise ValueError(message)


def check_region_geographic(region):
    """
    Check that the given geographic region is valid.

    A region is a bounding box for 2-dimensional coordinates (W, E, S, N).
    There should be 4 elements and lower latitude boundaries should not be
    larger than upper boundaries. Longitude boundaries should be in range [0,
    360] or [-180, 180] and latitude boundaries should be in range [-90, 90].

    Parameters
    ----------
    region : tuple = (W, E, S, N)
        The boundaries of a given region in geographic coordinates. Should have
        a lower and an upper boundary for each dimension of the coordinate
        system.

    Raises
    ------
    ValueError
        If the region doesn't have 4 entries, any lower boundary is larger than
        the upper boundary, or boundaries are outside their valid ranges.

    """
    if not region or len(region) != 4:
        message = (
            f"Invalid region '{region}'. Must have exactly 4 elements"
            " (W, E, S, N), a lower and an upper boundary for each dimension."
        )
        raise ValueError(message)
    west, east, south, north = region
    if south > north:
        message = f"Invalid region '{region}'. South boundary must be less than north."
        raise ValueError(message)
    if west < -180 or east > 360 or (west < 0 and east > 180):
        message = (
            f"Invalid region '{region}'. Longitude range must be [0, 360] "
            "or [-180, 180]."
        )
        raise ValueError(message)
    if south < -90 or north > 90:
        message = f"Invalid region '{region}'. Latitude range must be [-90, 90]."
        raise ValueError(message)


def check_adjust(adjust, valid=("spacing", "region")):
    """
    Check if the adjust argument is valid.

    Parameters
    ----------
    adjust : str
        The value of the adjust argument given to a function.
    valid : list or tuple
        The list of valid values for the argument.

    Raises
    ------
    ValueError
        In case the argument is not in the list of valid values.
    """
    if adjust not in valid:
        message = (
            f"Invalid value for 'adjust' argument '{adjust}'. Should be one of {valid}."
        )
        raise ValueError(message)


def check_shape(shape, region):
    """
    Check if the shape has a number of elements compatible with the region.

    The shape should have ``len(region) / 2`` elements. Assumes that the region
    is valid.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.
    shape : tuple = (..., size_SN, size_WE)
        The number of points in each direction of the given region, in reverse
        order. Must have one integer value per dimension of the region. The
        order of arguments is the opposite of the order of the region for
        compatibility with numpy's ``.shape`` attribute.

    Raises
    ------
    ValueError
        In case the number of elements in the shape is incorrect.
    """
    if not shape or not region or len(shape) != len(region) / 2:
        message = (
            f"Incompatible shape '{shape}' and region '{region}. "
            "There must be one element in 'shape' of every two elements in 'region'."
        )
        raise ValueError(message)


def check_overlap(overlap):
    """
    Check that the overlap argument is valid.

    It should be a decimal percentage in range [0, 1[. 100% overlap is not
    possible since that would generate infinite windows.

    Parameters
    ----------
    overlap : float
        The amount of overlap between adjacent windows. Should be within the
        range 1 > overlap ≥ 0.

    Raises
    ------
    ValueError
        In case the overlap is outside the allowed range.
    """
    if overlap < 0 or overlap >= 1:
        message = f"Invalid overlap '{overlap}'. Must be 1 > overlap >= 0."
        raise ValueError(message)


def longitude_continuity(region, *, coordinates=None):
    """
    Modify region boundaries and coordinates to ensure longitude continuity.

    Longitudinal boundaries of the region are moved to the ``[0, 360)`` or
    ``[-180, 180)`` degrees interval depending which one is better suited for
    that specific region to ensure that there are no discontinuities in the
    range. If given, coordinates are changed to match the new region.

    Longitude can be expressed as 0-360 or -180-180 and when specifying
    a small area, the longitude can have discontinuities in its range. For
    example, [350, 10] is a valid longitude range but wraps around at 0 instead
    of decreasing from 350 to 10. This can be problematic for some coordinate
    generation and manipulation routines.

    Parameters
    ----------
    region : tuple = (W, E, S, N)
        The boundaries of a given region in geographic coordinates. Should have
        a lower and an upper boundary for each dimension of the coordinate
        system.
    coordinates : tuple = (longitude, latitude) or None
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape. If None, we will assume that no
        coordinates are given.

    Returns
    -------
    modified_region : tuple = (W, E, S, N)
        The modified boundary of the region.
    modified_coordinates : tuple = (longitude, latitude)
        Only returned if coordinates are given. Modified set of geographic
        coordinates with continuous longitude.

    Examples
    --------
    Modify region with west > east to avoid the 360 discontinuity:

    >>> w, e, s, n = 350, 10, -10, 10
    >>> print(longitude_continuity([w, e, s, n]))
    (-10, 10, -10, 10)

    >>> w, e, s, n = 310, 180, -10, 10
    >>> print(longitude_continuity([w, e, s, n]))
    (-50, 180, -10, 10)

    For a global range of longitudes, always prefer the 0-360 range:

    >>> w, e, s, n = 0, 360, -10, 10
    >>> print(longitude_continuity([w, e, s, n]))
    (0, 360, -10, 10)

    >>> w, e, s, n = -180, 180, -10, 10
    >>> print(longitude_continuity([w, e, s, n]))
    (0, 360, -10, 10)

    Modify region and coordinates so that they match:

    >>> region = (350, 10, -60, -40)
    >>> coordinates = ((-10, -5, 0, 5, 10), (-60, -55, -50, -45, -40))
    >>> region, (longitude, latitude) = longitude_continuity(
    ...     region, coordinates=coordinates,
    ... )
    >>> print(region)
    (-10, 10, -60, -40)
    >>> print(longitude)
    [-10  -5   0   5  10]

    >>> region = (310, 180, -10, 10)
    >>> coordinates = ((310, 0, 180), (-10, 0, 10))
    >>> region, (longitude, latitude) = longitude_continuity(
    ...     region, coordinates=coordinates,
    ... )
    >>> print(region)
    (-50, 180, -10, 10)
    >>> print(longitude)
    [-50   0 180]

    The coordinates and region don't have to be within the same boundaries:

    >>> region = (-20, 20, -20, 20)
    >>> coordinates = ((0, 100, 200, 300), (-40, 0, 40, 80))
    >>> region, (longitude, latitude) = longitude_continuity(
    ...     region, coordinates=coordinates,
    ... )
    >>> print(region)
    (-20, 20, -20, 20)
    >>> print(longitude)
    [   0  100 -160  -60]
    """
    check_region_geographic(region)
    # Get longitudinal boundaries and check region
    w, e, s, n = region
    # Check if region is defined all around the globe
    all_globe = np.allclose(abs(e - w), 360)
    # Move coordinates to [0, 360)
    w = w % 360
    e = e % 360
    # Move west=0 and east=360 if region longitudes goes all around the globe
    if all_globe:
        w, e = 0, 360
    # Check if the [-180, 180) interval is better suited
    if w > e:
        interval_360 = False
        e = ((e + 180) % 360) - 180
        w = ((w + 180) % 360) - 180
        # If e = 180 then the above will create e=-180, which is wrong.
        # This only happens to the east limit and we need to fix it to have
        # a valid region.
        if e < w:
            e *= -1
    else:
        interval_360 = True
    modified_region = (w, e, s, n)
    if coordinates is not None:
        coordinates = check_coordinates(coordinates)
        check_coordinates_geographic(coordinates)
        # Run sanity checks for coordinates
        longitude = coordinates[0]
        longitude = longitude % 360
        if not interval_360:
            is_180 = np.isclose(longitude, 180)
            longitude = ((longitude + 180) % 360) - 180
            # Same as above when e < w leading to a -180 instead of 180.
            longitude[is_180] *= -1
        modified_coordinates = (longitude, *coordinates[1:])
        return modified_region, modified_coordinates
    return modified_region



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Generate regular coordinates along lines and grids.
"""

import numpy as np

from ._utils import spacing_to_size


def line_coordinates(
    start, stop, *, size=None, spacing=None, adjust="spacing", pixel_register=False
):
    """
    Generate evenly spaced points between two values.

    Able to handle either specifying the number of points required (*size*) or
    the size of the interval between points (*spacing*). If using *size*, the
    output will be similar to using :func:`numpy.linspace`. When using
    *spacing*, if the interval is not divisible by the desired spacing, either
    the interval or the spacing will have to be adjusted. By default, the
    spacing will be rounded to the nearest multiple. Optionally, the *start*
    and *stop* values (the region) can be adjusted to fit the exact spacing
    given.

    Parameters
    ----------
    start : float
        The starting value of the sequence.
    stop : float
        The end value of the sequence.
    size : int or None
        The number of points in the sequence. If None, *spacing* must be
        provided.
    spacing : float or None
        The step size (interval) between points in the sequence. If None,
        *size* must be provided.
    adjust : str = "spacing" or "region"
        Whether to adjust the spacing or the interval/region if required.
        Ignored if *size* is given instead of *spacing*. Default is
        ``"spacing"``
    pixel_register : bool
        If True, the points will refer to the center of each interval (pixel)
        instead of the boundaries. In practice, this means that there will be
        one less element in the sequence if *spacing* is provided. If *size* is
        provided, the requested number of elements is respected. Default is
        False.

    Returns
    -------
    sequence : array
        The generated sequence of values.

    Examples
    --------
    >>> values = line_coordinates(0, 5, spacing=2.5)
    >>> print(values.shape)
    (3,)
    >>> print(values)
    [0.  2.5 5. ]
    >>> print(line_coordinates(0, 10, size=5))
    [ 0.   2.5  5.   7.5 10. ]
    >>> print(line_coordinates(0, 10, spacing=2.5))
    [ 0.   2.5  5.   7.5 10. ]

    The spacing is adjusted to fit the interval by default but this can be
    changed to adjusting the interval/region instead:

    >>> print(line_coordinates(0, 10, spacing=2.4))
    [ 0.   2.5  5.   7.5 10. ]
    >>> print(line_coordinates(0, 10, spacing=2.4, adjust="region"))
    [0.2 2.6 5.  7.4 9.8]
    >>> print(line_coordinates(0, 10, spacing=2.6))
    [ 0.   2.5  5.   7.5 10. ]
    >>> print(line_coordinates(0, 10, spacing=2.6, adjust="region"))
    [-0.2  2.4  5.   7.6 10.2]

    Optionally, return values at the center of the intervals instead of their
    boundaries:

    >>> print(line_coordinates(0, 10, spacing=2.5, pixel_register=True))
    [1.25 3.75 6.25 8.75]

    Notice that this produces one value less than the non-pixel registered
    version. If using *size* instead of *spacing*, the number of values will be
    *size* regardless and the spacing will therefore be different from the
    non-pixel registered version:

    >>> print(line_coordinates(0, 10, size=5, pixel_register=True))
    [1. 3. 5. 7. 9.]

    """
    if size is not None and spacing is not None:
        message = "Both size and spacing provided. Only one is allowed."
        raise ValueError(message)
    if size is None and spacing is None:
        message = "Either a size or a spacing must be provided."
        raise ValueError(message)
    if spacing is not None:
        size, start, stop = spacing_to_size(start, stop, spacing, adjust=adjust)
    elif pixel_register and size is not None:
        # Starts by generating grid-line registered coordinates and shifting
        # them to the center of the pixel. Need 1 more point if given a size
        # instead of spacing so that we can do that because we discard the last
        # point when shifting the coordinates.
        size = size + 1
    if size < 1:
        message = f"Invalid size '{size}'. Must be >= 1."
        raise ValueError(message)
    values = np.linspace(start, stop, size)
    if pixel_register:
        values = values[:-1] + (values[1] - values[0]) / 2
    return values



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions for validating and manipulation coordinate arrays.
"""

import numpy as np

from ._validation import check_adjust, check_region, check_shape


def spacing_to_size(start, stop, spacing, *, adjust="spacing"):
    """
    Convert a spacing to the number of points between start and stop.

    Takes into account if the spacing or the interval needs to be adjusted in
    order to fit exactly. This is needed when the interval is not a multiple of
    the spacing.

    Parameters
    ----------
    start : float
        The starting value of the sequence.
    stop : float
        The end value of the sequence.
    spacing : float
        The step size (interval) between points in the sequence.
    adjust : {'spacing', 'region'}
        Whether to adjust the spacing or the interval/region if required.
        Defaults to adjusting the spacing.

    Returns
    -------
    size : int
        The number of points between start and stop.
    start : float
        The end of the interval, which may or may not have been adjusted.
    stop : float
        The end of the interval, which may or may not have been adjusted.

    Examples
    --------
    If the spacing is a multiple of the interval, then the size is how many
    points fit in the interval and the start and stop values are maintained:

    >>> size, start, stop = spacing_to_size(0, 1, 0.5)
    >>> print(size, start, stop)
    3 0 1

    If the spacing is not a multiple, then it will be adjusted to fit the
    interval by default. In this case, then number of points remains the same:

    >>> size, start, stop = spacing_to_size(0, 1, 0.6)
    >>> print(size, start, stop)
    3 0 1

    Alternatively, we can ask it to adjust the region instead of the spacing
    between points:

    >>> size, start, stop = spacing_to_size(0, 1, 0.6, adjust="region")
    >>> print(f"{size} {start:.1f} {stop:.1f}")
    3 -0.1 1.1

    If the start and stop are the same, only a single point will be generated:

    >>> size, start, stop = spacing_to_size(1e-8, 1e-8, 1e-7)
    >>> print(f"{size} {start:.1g} {stop:.1g}")
    1 1e-08 1e-08
    """
    check_adjust(adjust)
    # Add 1 to get the number of nodes, not segments
    size = int(round((stop - start) / spacing) + 1)
    # If the spacing >= 2 * (stop - start), it rounds to zero so we'd be
    # generating a single point, which isn't equivalent to adjusting the
    # spacing or the region. To get the appropriate behaviour of decreasing the
    # spacing until it fits the region or increasing the region until it fits
    # at least 1 spacing, we need to always round to at least 1 in the code
    # above. But this should only be done if the interval stop - start is not
    # close to zero.
    if size == 1 and not np.isclose(stop, start):
        # if size == 1:
        size += 1
    if adjust == "region":
        # The size is the same but we adjust the interval so that the spacing
        # isn't altered when we do the linspace.
        required_length = (size - 1) * spacing
        given_length = stop - start
        pad = (required_length - given_length) / 2
        stop = stop + pad
        start = start - pad
    return size, start, stop


def shape_to_spacing(region, shape, *, pixel_register=False):
    """
    Calculate the spacing of a regular grid given a region and shape.

    The spacing is assumed to be constant along each direction but can vary
    between directions.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.
    shape : tuple = (..., size_SN, size_WE)
        The number of points in each direction of the given region, in reverse
        order. Must have one integer value per dimension of the region. The
        order of arguments is the opposite of the order of the region for
        compatibility with numpy's ``.shape`` attribute.
    pixel_register : bool
        If True, the coordinates will refer to the center of each grid pixel
        instead of the grid lines. In practice, this means that there will be
        one less element per dimension of the grid when compared to grid line
        registered. Default is False.

    Returns
    -------
    spacing : tuple = (..., space_SN, space_WE)
        The grid spacing in each direction of the given region, in reverse
        order. Has one value per dimension of the region. The order of
        arguments is the opposite of the order of the region for compatibility
        with *shape*.

    Examples
    --------
    For 2-dimensional grids, the region has 4 elements and the shape must have
    2 elements:

    >>> spacing = shape_to_spacing((0, 10, -5, 1), (7, 11))
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}")
    1.0, 1.0

    The spacing doesn't have to be the same in each direction:

    >>> spacing = shape_to_spacing((0, 10, -5, 1), (14, 11))
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}")
    0.5, 1.0

    Notice that the spacing is in the same order as the shape:

    >>> spacing = shape_to_spacing((0, 10, -5, 1), (7, 21))
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}")
    1.0, 0.5

    Pixel registration is also supported:

    >>> spacing = shape_to_spacing(
    ...     (-0.5, 10.5, -5.5, 1.5), (7, 11), pixel_register=True,
    ... )
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}")
    1.0, 1.0
    >>> spacing = shape_to_spacing(
    ...     (-0.25, 10.25, -5.5, 1.5), (7, 21), pixel_register=True,
    ... )
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}")
    1.0, 0.5

    Grids don't have to be 2-dimensional:

    >>> spacing = shape_to_spacing((0, 10, -5, 1, 10, 14), (5, 7, 11))
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}, {spacing[2]:.1f}")
    1.0, 1.0, 1.0
    >>> spacing = shape_to_spacing(
    ...     (-0.25, 10.25, -5.5, 1.5, -0.1, 1.1), (5, 7, 21), pixel_register=True,
    ... )
    >>> print(f"{spacing[0]:.1f}, {spacing[1]:.1f}, {spacing[2]:.1f}")
    0.2, 1.0, 0.5
    """
    check_region(region)
    check_shape(shape, region)
    spacing = []
    for i, n in enumerate(reversed(shape)):
        n_points = n
        if not pixel_register:
            n_points -= 1
        spacing.append((region[2 * i + 1] - region[2 * i]) / n_points)
    return tuple(reversed(spacing))


def make_non_dimensional_coordinates(values, shape, dtype):
    """
    Make a list of arrays with the given values of non-dimensional coordinates.

    These are coordinates that are not part of the dimensions of a grid or
    spread of points. Think of them as extra coordinates.

    Parameters
    ----------
    values : None or float or list or array
        The values of the non-dimensional coordinates. If None, an empty list
        will be returned. If a float, will generate one array of the desired
        shape with this value as the fill. If a list or array, will generate
        one array of the given shape per value in the list or array.
    shape : tuple
        The shape of the desired array(s).
    dtype : str
        A numpy compatible dtype used for the output array.

    Returns
    -------
    coordinates : list
        List of arrays generated. Even if a single value is passed, the output
        will still be a list with a single array.

    Examples
    --------
    Generating a single array:

    >>> make_non_dimensional_coordinates(10, (2,), dtype="int")
    [array([10, 10])]
    >>> make_non_dimensional_coordinates(10, (2, 2), dtype="int")
    [array([[10, 10],
           [10, 10]])]

    To generate multiple arrays:

    >>> make_non_dimensional_coordinates([10, 20], (2,), dtype="int")
    [array([10, 10]), array([20, 20])]
    >>> import numpy as np
    >>> make_non_dimensional_coordinates(np.array([10, 20]), (2,), dtype="int")
    [array([10, 10]), array([20, 20])]
    >>> make_non_dimensional_coordinates(
    ...     np.array([[10], [20]]), (2,), dtype="int",
    ... )
    [array([10, 10]), array([20, 20])]

    If None is given:

    >>> make_non_dimensional_coordinates(None, (2,), dtype="int")
    []
    """
    coordinates = []
    if values is not None:
        for value in np.atleast_1d(values).ravel():
            coordinates.append(np.full(shape, value, dtype=dtype))
    return coordinates



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Generate coordinates for regular n-dimensional grids.
"""

import numpy as np

from ._line import line_coordinates
from ._utils import make_non_dimensional_coordinates
from ._validation import check_region, check_shape


def grid_coordinates(
    region,
    *,
    shape=None,
    spacing=None,
    adjust="spacing",
    pixel_register=False,
    non_dimensional_coords=None,
):
    """
    Generate evenly spaced points on an n-dimensional grid.

    The grid can be specified by either the number of points in each dimension
    (the *shape*) or by the grid node *spacing* in each dimension.

    If the given region is not divisible by the desired spacing, either the
    region or the spacing will have to be adjusted. By default, the spacing
    will be rounded to the nearest multiple. Optionally, the boundaries of the
    region can be adjusted to fit the exact spacing given. See the examples
    below.

    Supports laying out points on a grid-node registration or a pixel
    registration. See examples below for more information.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.
    shape : tuple = (..., size_SN, size_WE) or None
        The number of points in each direction of the given region, in reverse
        order. Must have one integer value per dimension of the region. The
        order of arguments is the opposite of the order of the region for
        compatibility with numpy's ``.shape`` attribute. If None, *spacing*
        must be provided. Default is None.
    spacing : float, tuple = (..., space_SN, space_WE), or None
        The grid spacing in each direction of the given region, in reverse
        order. A single value means that the spacing is equal in all
        directions. If a tuple, must have one value per dimension of the
        region. The order of arguments is the opposite of the order of the
        region for compatibility with *shape*. If None, *shape* must be
        provided. Default is None.
    adjust : str = "spacing" or "region"
        Whether to adjust the spacing or the region if the spacing is not
        a multiple of the region. Ignored if *shape* is given instead of
        *spacing*. Default is ``"spacing"``.
    pixel_register : bool
        If True, the coordinates will refer to the center of each grid pixel
        instead of the grid lines. In practice, this means that there will be
        one less element per dimension of the grid when compared to grid line
        registered (only if given *spacing* and not *shape*). Default is False.
    non_dimensional_coords : None, scalar, or tuple of scalars
        If not None, then value(s) of extra non-dimensional coordinates
        (coordinates that aren't part of the grid dimensions, like height for
        a lat/lon grid). Will generate extra coordinate arrays from these
        values with the same shape of the final grid coordinates and the
        constant value given here. Use this to generate arrays of constant
        heights or times, for example, which might be needed to accompany
        a grid.

    Returns
    -------
    coordinates : tuple of arrays
        Arrays with coordinates of each point in the grid. Each array contains
        values for a dimension in the order of the given region, and any extra
        values given in *non_dimensional_coords*. All arrays will have the same
        shape.

    Examples
    --------
    >>> east, north = grid_coordinates(region=(0, 5, 0, 10), shape=(5, 3))
    >>> print(east.shape, north.shape)
    (5, 3) (5, 3)
    >>> print(east)
    [[0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]]
    >>> print(north)
    [[ 0.   0.   0. ]
     [ 2.5  2.5  2.5]
     [ 5.   5.   5. ]
     [ 7.5  7.5  7.5]
     [10.  10.  10. ]]

    The grid can also be specified using the spacing between points instead of
    the shape:

    >>> east, north = grid_coordinates(region=(0, 5, 0, 10), spacing=2.5)
    >>> print(east.shape, north.shape)
    (5, 3) (5, 3)
    >>> print(east)
    [[0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]
     [0.  2.5 5. ]]
    >>> print(north)
    [[ 0.   0.   0. ]
     [ 2.5  2.5  2.5]
     [ 5.   5.   5. ]
     [ 7.5  7.5  7.5]
     [10.  10.  10. ]]

    The spacing can be different for northing and easting, respectively:

    >>> east, north = grid_coordinates(region=(-5, 1, 0, 10), spacing=(2.5, 1))
    >>> print(east.shape, north.shape)
    (5, 7) (5, 7)
    >>> print(east)
    [[-5. -4. -3. -2. -1.  0.  1.]
     [-5. -4. -3. -2. -1.  0.  1.]
     [-5. -4. -3. -2. -1.  0.  1.]
     [-5. -4. -3. -2. -1.  0.  1.]
     [-5. -4. -3. -2. -1.  0.  1.]]
    >>> print(north)
    [[ 0.   0.   0.   0.   0.   0.   0. ]
     [ 2.5  2.5  2.5  2.5  2.5  2.5  2.5]
     [ 5.   5.   5.   5.   5.   5.   5. ]
     [ 7.5  7.5  7.5  7.5  7.5  7.5  7.5]
     [10.  10.  10.  10.  10.  10.  10. ]]

    Grids of more dimensions can be generated by specifying more elements in
    ``region`` and ``shape``/``spacing``:

    >>> east, north, up = grid_coordinates(
    ...     region=(0, 5, 0, 10, -4, 0), shape=(2, 5, 3)
    ... )
    >>> print(east.shape, north.shape, up.shape)
    (2, 5, 3) (2, 5, 3) (2, 5, 3)
    >>> print(east)
    [[[0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]]
    <BLANKLINE>
     [[0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]
      [0.  2.5 5. ]]]
    >>> print(north)
    [[[ 0.   0.   0. ]
      [ 2.5  2.5  2.5]
      [ 5.   5.   5. ]
      [ 7.5  7.5  7.5]
      [10.  10.  10. ]]
    <BLANKLINE>
     [[ 0.   0.   0. ]
      [ 2.5  2.5  2.5]
      [ 5.   5.   5. ]
      [ 7.5  7.5  7.5]
      [10.  10.  10. ]]]
    >>> print(up)
    [[[-4. -4. -4.]
      [-4. -4. -4.]
      [-4. -4. -4.]
      [-4. -4. -4.]
      [-4. -4. -4.]]
    <BLANKLINE>
     [[ 0.  0.  0.]
      [ 0.  0.  0.]
      [ 0.  0.  0.]
      [ 0.  0.  0.]
      [ 0.  0.  0.]]]

    If the region can't be divided into the desired spacing, the spacing will
    be adjusted to conform to the region:

    >>> east, north = grid_coordinates(region=(-5, 0, 0, 5), spacing=2.6)
    >>> print(east.shape, north.shape)
    (3, 3) (3, 3)
    >>> print(east)
    [[-5.  -2.5  0. ]
     [-5.  -2.5  0. ]
     [-5.  -2.5  0. ]]
    >>> print(north)
    [[0.  0.  0. ]
     [2.5 2.5 2.5]
     [5.  5.  5. ]]
    >>> east, north = grid_coordinates(region=(-5, 0, 0, 5), spacing=2.4)
    >>> print(east.shape, north.shape)
    (3, 3) (3, 3)
    >>> print(east)
    [[-5.  -2.5  0. ]
     [-5.  -2.5  0. ]
     [-5.  -2.5  0. ]]
    >>> print(north)
    [[0.  0.  0. ]
     [2.5 2.5 2.5]
     [5.  5.  5. ]]

    You can choose to adjust the East and North boundaries of the region
    instead:

    >>> east, north = grid_coordinates(
    ...     region=(-5, 0, 0, 5), spacing=2.6, adjust='region',
    ... )
    >>> print(east.shape, north.shape)
    (3, 3) (3, 3)
    >>> print(east)
    [[-5.1 -2.5  0.1]
     [-5.1 -2.5  0.1]
     [-5.1 -2.5  0.1]]
    >>> print(north)
    [[-0.1 -0.1 -0.1]
     [ 2.5  2.5  2.5]
     [ 5.1  5.1  5.1]]
    >>> east, north = grid_coordinates(
    ...     region=(-5, 0, 0, 5), spacing=2.4, adjust='region',
    ... )
    >>> print(east.shape, north.shape)
    (3, 3) (3, 3)
    >>> print(east)
    [[-4.9 -2.5 -0.1]
     [-4.9 -2.5 -0.1]
     [-4.9 -2.5 -0.1]]
    >>> print(north)
    [[0.1 0.1 0.1]
     [2.5 2.5 2.5]
     [4.9 4.9 4.9]]

    We can optionally generate coordinates for the center of each grid pixel
    instead of the corner (default):

    >>> east, north = grid_coordinates(
    ...     region=(0, 5, 0, 10), spacing=2.5, pixel_register=True,
    ... )
    >>> # Notice that the shape is 1 less than when pixel_register=False
    >>> print(east.shape, north.shape)
    (4, 2) (4, 2)
    >>> print(east)
    [[1.25 3.75]
     [1.25 3.75]
     [1.25 3.75]
     [1.25 3.75]]
    >>> print(north)
    [[1.25 1.25]
     [3.75 3.75]
     [6.25 6.25]
     [8.75 8.75]]
    >>> east, north = grid_coordinates(
    ...     region=(0, 5, 0, 10), shape=(4, 2), pixel_register=True,
    ... )
    >>> print(east)
    [[1.25 3.75]
     [1.25 3.75]
     [1.25 3.75]
     [1.25 3.75]]
    >>> print(north)
    [[1.25 1.25]
     [3.75 3.75]
     [6.25 6.25]
     [8.75 8.75]]

    Generate arrays for other coordinates that have a constant value:

    >>> east, north, height = grid_coordinates(
    ...     region=(0, 5, 0, 10), spacing=2.5, non_dimensional_coords=57
    ... )
    >>> print(east.shape, north.shape, height.shape)
    (5, 3) (5, 3) (5, 3)
    >>> print(height)
    [[57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]]
    >>> east, north, height, time = grid_coordinates(
    ...     region=(0, 5, 0, 10), spacing=2.5, non_dimensional_coords=[57, 0.1]
    ... )
    >>> print(east.shape, north.shape, height.shape, time.shape)
    (5, 3) (5, 3) (5, 3) (5, 3)
    >>> print(height)
    [[57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]
     [57. 57. 57.]]
    >>> print(time)
    [[0.1 0.1 0.1]
     [0.1 0.1 0.1]
     [0.1 0.1 0.1]
     [0.1 0.1 0.1]
     [0.1 0.1 0.1]]

    """
    check_region(region)
    ndims = len(region) // 2
    if shape is not None and spacing is not None:
        message = (
            f"Both grid shape ('{shape}') and spacing ('{spacing}') were provided. "
            "Only one is allowed."
        )
        raise ValueError(message)
    if shape is None and spacing is None:
        message = "Either a grid shape or a spacing must be provided."
        raise ValueError(message)
    if shape is None:
        shape = tuple(None for _ in range(ndims))
        if np.isscalar(spacing):
            spacing = tuple(spacing for _ in range(ndims))
        elif len(spacing) != ndims:
            message = (
                f"Invalid spacing '{spacing}'. Should have {ndims} values, "
                f"one value per dimension of the region '{region}'."
            )
            raise ValueError(message)
    else:
        spacing = tuple(None for _ in range(ndims))
    # Must check the shape after so that it's guaranteed to be a tuple.
    check_shape(shape, region)
    region_pairs = np.reshape(region, (len(region) // 2, 2))
    coordinates_1d = []
    for size, space, (low, up) in zip(reversed(shape), reversed(spacing), region_pairs):
        coordinates_1d.append(
            line_coordinates(
                low,
                up,
                size=size,
                spacing=space,
                adjust=adjust,
                pixel_register=pixel_register,
            )
        )
    # Use the ij indexing in meshgrid because the default xy works for 2D
    # arrays fine but mixes the order of coordinates for 3D and beyond. But
    # this means we have to reverse the list going in to get the right shape
    # and reverse it coming out to get the right order of coordinates.
    coordinates = list(reversed(np.meshgrid(*reversed(coordinates_1d), indexing="ij")))
    coordinates.extend(
        make_non_dimensional_coordinates(
            non_dimensional_coords, coordinates[0].shape, coordinates[0].dtype
        )
    )
    return tuple(coordinates)



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions for generating random point spreads.
"""

import numpy as np

from ._utils import make_non_dimensional_coordinates
from ._validation import check_region, check_region_geographic, longitude_continuity


def random_coordinates(region, size, *, random_seed=None, non_dimensional_coords=None):
    """
    Generate the coordinates for a uniformly random scatter of points.

    The points are drawn from a uniform distribution, independently for each
    dimension of the given region.

    Parameters
    ----------
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. Should have a lower and an upper boundary for each
        dimension of the coordinate system.
    size : int
        The number of points to generate.
    random_seed : None or int or numpy.random.Generator
        A seed for a random number generator (RNG) used to generate the
        coordinates. If an integer is given, it will be used as a seed for
        :func:`numpy.random.default_rng` which will then be used as the
        generator. If a :class:`numpy.random.Generator` is given, it will be
        used. If ``None`` is given, :func:`~numpy.random.default_rng` will be
        used with no seed to create a generator (resulting in different numbers
        with each run). Use a seed to make sure computations are reproducible.
        Default is None.
    non_dimensional_coords : None, scalar, or tuple of scalars
        If not None, then value(s) of extra non-dimensional coordinates
        (coordinates that aren't part of the sample dimensions, like height for
        a lat/lon grid). Will generate extra coordinate arrays from these
        values with the same shape of the final coordinates and the constant
        value given here. Use this to generate arrays of constant heights or
        times, for example, which might be needed to accompany a set of points.

    Returns
    -------
    coordinates : tuple of arrays
        Arrays with coordinates of each point in the grid. Each array contains
        values for a dimension in an order compatible with *region* followed by
        any extra dimensions given in *non_dimensional_coords*. All arrays will
        have the specified *size*.

    Examples
    --------
    We'll use a seed value to ensure that the same will be generated every
    time:

    >>> easting, northing = random_coordinates(
    ...     (0, 10, -2, -1), size=4, random_seed=0,
    ... )
    >>> print(', '.join(['{:.4f}'.format(i) for i in easting]))
    6.3696, 2.6979, 0.4097, 0.1653
    >>> print(', '.join(['{:.4f}'.format(i) for i in northing]))
    -1.1867, -1.0872, -1.3934, -1.2705
    >>> easting, northing, height = random_coordinates(
    ...     (0, 10, -2, -1), 4, random_seed=0, non_dimensional_coords=12
    ... )
    >>> print(height)
    [12. 12. 12. 12.]
    >>> easting, northing, height, time = random_coordinates(
    ...     (0, 10, -2, -1),
    ...     size=4,
    ...     random_seed=0,
    ...     non_dimensional_coords=[12, 1986],
    ... )
    >>> print(height)
    [12. 12. 12. 12.]
    >>> print(time)
    [1986. 1986. 1986. 1986.]

    We're not limited to 2 dimensions:

    >>> easting, northing, up = random_coordinates(
    ...     (0, 10, -2, -1, 0.1, 0.2), 4, random_seed=0,
    ... )
    >>> print(', '.join(['{:.4f}'.format(i) for i in easting]))
    6.3696, 2.6979, 0.4097, 0.1653
    >>> print(', '.join(['{:.4f}'.format(i) for i in northing]))
    -1.1867, -1.0872, -1.3934, -1.2705
    >>> print(', '.join(['{:.4f}'.format(i) for i in up]))
    0.1544, 0.1935, 0.1816, 0.1003

    """
    check_region(region)
    random = np.random.default_rng(random_seed)
    coordinates = []
    for lower, upper in np.reshape(region, (len(region) // 2, 2)):
        coordinates.append(random.uniform(lower, upper, size))
    coordinates.extend(
        make_non_dimensional_coordinates(
            non_dimensional_coords, coordinates[0].shape, coordinates[0].dtype
        )
    )
    return tuple(coordinates)


def random_coordinates_spherical(
    region, size, *, random_seed=None, non_dimensional_coords=None
):
    """
    Generate the coordinates for uniformly random points on the sphere.

    Points drawn from a simple uniform distribution of longitude and latitude
    will tend to be more concentrated towards the poles. This function accounts
    for that and is able generate a uniformly random distribution on the
    surface of a sphere.

    Parameters
    ----------
    region : tuple = (W, E, S, N)
        The boundaries of a given region in geographic coordinates. Should have
        a lower and an upper boundary for each dimension of the coordinate
        system.
    size : int
        The number of points to generate.
    random_seed : None or int or numpy.random.Generator
        A seed for a random number generator (RNG) used to generate the
        coordinates. If an integer is given, it will be used as a seed for
        :func:`numpy.random.default_rng` which will then be used as the
        generator. If a :class:`numpy.random.Generator` is given, it will be
        used. If ``None`` is given, :func:`~numpy.random.default_rng` will be
        used with no seed to create a generator (resulting in different numbers
        with each run). Use a seed to make sure computations are reproducible.
        Default is None.
    non_dimensional_coords : None, scalar, or tuple of scalars
        If not None, then value(s) of extra non-dimensional coordinates
        (coordinates that aren't part of the sample dimensions, like height for
        a lat/lon grid). Will generate extra coordinate arrays from these
        values with the same shape of the final coordinates and the constant
        value given here. Use this to generate arrays of constant heights or
        times, for example, which might be needed to accompany a set of points.

    Returns
    -------
    coordinates : tuple of arrays
        Arrays with the longitude, latitude, and non-dimensional coordinates,
        in order, of each point in the grid. Each array contains values for
        a dimension in an order compatible with *region* followed by any extra
        dimensions given in *non_dimensional_coords*. All arrays will have the
        specified *size*.

    Examples
    --------
    We can generate the random coordinates on a sphere like so:

    >>> coordinates = random_coordinates_spherical(
    ...     region=(-100, 100, -80, -20), size=10, random_seed=42,
    ... )

    We set a seed here to make sure our examples always return the same values.
    If you need different values every time you run your code, then either omit
    ``random_seed`` or set it to ``None``.

    >>> import numpy as np

    The first coordinate is the longitude:

    >>> print(np.array_str(coordinates[0], precision=1))
    [ 54.8 -12.2  71.7  39.5 -81.2  95.1  52.2  57.2 -74.4  -9.9]

    And the second is the latitude:

    >>> print(np.array_str(coordinates[1], precision=1))
    [-48.3 -22.9 -34.8 -27.1 -44.4 -57.  -38.9 -70.7 -26.9 -35.4]

    To show how this differs from :func:`bordado.random_coordinates`, we can
    generate a large number of points and calculate the point density per
    latitude band. We expect the concentration to be uniform since we want
    uniformly distributed numbers.

    First, we'll define a function that calculates the point density per 10
    degree band of latitude:

    >>> import bordado as bd
    >>> def point_density(coordinates, region):
    ...     # Define the latitude bands
    ...     bands = bd.line_coordinates(*region[2:], spacing=10)
    ...     # Calculate the area of each band.
    ...     # See https://en.wikipedia.org/wiki/Spherical_cap
    ...     areas = 2 * np.pi * abs(
    ...         np.sin(np.radians(bands[:-1])) - np.sin(np.radians(bands[1:]))
    ...     )
    ...     # Figure out how many points are in each band
    ...     points_per_band = np.array([
    ...         bd.inside(
    ...             coordinates,
    ...             [*region[:2], bands[i], bands[i + 1]],
    ...         ).sum()
    ...         for i in range(bands.size - 1)
    ...     ])
    ...     # Calculate the density
    ...     density = points_per_band / areas
    ...     return density

    Now we can make a lot of random points using this function and the
    traditional :func:`bordado.random_coordinates` to compare:

    >>> region = (0, 360, -90, 90)
    >>> size = 100_000
    >>> coordinates_cartesian = random_coordinates(
    ...     region, size, random_seed=42,
    ... )
    >>> coordinates_spherical = random_coordinates_spherical(
    ...     region, size, random_seed=42,
    ... )

    Finally, we calculate and print the point density per latitude for each
    set:

    >>> density_cartesian = point_density(coordinates_cartesian, region)
    >>> print(np.array_str(density_cartesian, precision=0))
    [57147. 19812. 12159.  8981.  7267.  6107.  5574.  5320.  5151.  5033.
      5237.  5558.  6138.  7075.  8800. 11967. 19660. 58635.]

    >>> density_spherical = point_density(coordinates_spherical, region)
    >>> print(np.array_str(density_spherical, precision=0))
    [8192. 7786. 7683. 8134. 8113. 8112. 7867. 7982. 8029. 7945. 7834. 7901.
     7931. 7902. 7924. 8037. 7987. 8119.]

    As you can see, for the Cartesian version the density increases towards the
    poles but for the spherical version it stays roughly the same throughout
    the globe.
    """
    check_region_geographic(region)
    region = longitude_continuity(region)
    random = np.random.default_rng(random_seed)
    colat_south = np.radians(90 - region[2])
    colat_north = np.radians(90 - region[3])
    xmax = (1 + np.cos(colat_north)) / 2
    xmin = (1 + np.cos(colat_south)) / 2
    coordinates = [
        random.uniform(*region[:2], size),
        90 - np.degrees(np.arccos(2 * random.uniform(xmin, xmax, size) - 1)),
    ]
    coordinates.extend(
        make_non_dimensional_coordinates(
            non_dimensional_coords, coordinates[0].shape, coordinates[0].dtype
        )
    )
    return tuple(coordinates)



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions to split points into blocks and windows.
"""

import numpy as np
from scipy.spatial import KDTree

from ._grid import grid_coordinates, line_coordinates
from ._region import get_region, pad_region
from ._validation import (
    check_adjust,
    check_coordinates,
    check_coordinates_geographic,
    check_overlap,
    check_region,
    check_region_geographic,
    longitude_continuity,
)


def block_split(
    coordinates, *, region=None, block_shape=None, block_size=None, adjust="block_size"
):
    """
    Split a region into blocks and label points according to where they fall.

    The labels are integers corresponding to the index of the block. Also
    returns the coordinates of the center of each block (following the same
    index as the labels). Blocks can be specified by their size or the number
    of blocks in each dimension (the shape).

    Uses :class:`scipy.spatial.KDTree` to nearest neighbor lookup during the
    splitting process.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape.
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. If region is not given, will use the bounding region of
        the given coordinates.
    block_shape : tuple = (..., n_north, n_east) or None
        The number of blocks in each direction, in reverse order. Must have one
        integer value per coordinate dimension. The order of arguments is the
        opposite of the order of the region for compatibility with numpy's
        ``.shape`` attribute. If None, *block_size* must be provided. Default
        is None.
    block_size : float, tuple = (..., size_north, size_east), or None
        The block size in each direction, in reverse order. A single value
        means that the block size is equal in all directions. If a tuple, must
        have one value per dimension of the coordinates. The order of arguments
        is the opposite of the order of the coordinates for compatibility with
        *block_shape*. If None, *block_shape* must be provided. Default is
        None.
    adjust : str = "block_size" or "region"
        Whether to adjust the block size or the region, if required. Adjusting
        the size or region is required when the block size is not a multiple of
        the region. Ignored if *block_shape* is given instead of *block_size*.
        Defaults to adjusting the block size.

    Returns
    -------
    block_coordinates : tuple = (easting, northing, ...)
        ND arrays with the coordinates of the center of each block.
    labels : array
        Array with the same shape as the block coordinates. Contains the
        integer label for each data point. The label is the index of the block
        to which that point belongs.

    Examples
    --------
    Let's make some points along a 2D grid to try splitting (the points don't
    have to be on a grid but this makes it easier to explain):

    >>> import bordado as bd
    >>> coordinates = bd.grid_coordinates((-5, 0, 5, 10), spacing=1)
    >>> print(coordinates[0].shape)
    (6, 6)
    >>> print(coordinates[0])
    [[-5. -4. -3. -2. -1.  0.]
     [-5. -4. -3. -2. -1.  0.]
     [-5. -4. -3. -2. -1.  0.]
     [-5. -4. -3. -2. -1.  0.]
     [-5. -4. -3. -2. -1.  0.]
     [-5. -4. -3. -2. -1.  0.]]
    >>> print(coordinates[1])
    [[ 5.  5.  5.  5.  5.  5.]
     [ 6.  6.  6.  6.  6.  6.]
     [ 7.  7.  7.  7.  7.  7.]
     [ 8.  8.  8.  8.  8.  8.]
     [ 9.  9.  9.  9.  9.  9.]
     [10. 10. 10. 10. 10. 10.]]

    We can split into blocks by specifying the block size:

    >>> block_coords, labels = block_split(coordinates, block_size=2.5)

    The first argument is a tuple of coordinates for the center of each block:

    >>> print(len(block_coords))
    2
    >>> print(block_coords[0])
    [[-3.75 -1.25]
     [-3.75 -1.25]]
    >>> print(block_coords[1])
    [[6.25 6.25]
     [8.75 8.75]]

    The labels are an array of the same shape as the coordinates and has the
    index of the block each point belongs to:

    >>> print(labels)
    [[0 0 0 1 1 1]
     [0 0 0 1 1 1]
     [0 0 0 1 1 1]
     [2 2 2 3 3 3]
     [2 2 2 3 3 3]
     [2 2 2 3 3 3]]

    Use this to index the coordinates, for example to get all points that fall
    inside the first block:

    >>> block_0 = [c[labels == 0] for c in coordinates]
    >>> print(block_0[0])
    [-5. -4. -3. -5. -4. -3. -5. -4. -3.]
    >>> print(block_0[1])
    [5. 5. 5. 6. 6. 6. 7. 7. 7.]

    You can also specify the number of blocks along each direction instead of
    their size:

    >>> block_coords, labels = block_split(coordinates, block_shape=(4, 2))
    >>> print(len(block_coords))
    2
    >>> print(block_coords[0])
    [[-3.75 -1.25]
     [-3.75 -1.25]
     [-3.75 -1.25]
     [-3.75 -1.25]]
    >>> print(block_coords[1])
    [[5.625 5.625]
     [6.875 6.875]
     [8.125 8.125]
     [9.375 9.375]]
    >>> print(labels)
    [[0 0 0 1 1 1]
     [0 0 0 1 1 1]
     [2 2 2 3 3 3]
     [4 4 4 5 5 5]
     [6 6 6 7 7 7]
     [6 6 6 7 7 7]]

    By default, the region (bounding box of the points) will be derived from
    the coordinates. You can also specify a custom region for the splitting if
    desired:

    >>> block_coords, labels = block_split(
    ...     coordinates, block_size=2, region=(-5.5, 0.5, 4.5, 10.5),
    ... )
    >>> print(block_coords[0])
    [[-4.5 -2.5 -0.5]
     [-4.5 -2.5 -0.5]
     [-4.5 -2.5 -0.5]]
    >>> print(block_coords[1])
    [[5.5 5.5 5.5]
     [7.5 7.5 7.5]
     [9.5 9.5 9.5]]
    >>> print(labels)
    [[0 0 1 1 2 2]
     [0 0 1 1 2 2]
     [3 3 4 4 5 5]
     [3 3 4 4 5 5]
     [6 6 7 7 8 8]
     [6 6 7 7 8 8]]

    Coordinates can be more than 2-dimensional as well:

    >>> coordinates = bd.grid_coordinates((-5, 0, 5, 10, 1, 2), spacing=1)
    >>> print(coordinates[0].shape)
    (2, 6, 6)
    >>> print(coordinates[0])
    [[[-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]]
    <BLANKLINE>
     [[-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]]]
    >>> print(coordinates[1])
    [[[ 5.  5.  5.  5.  5.  5.]
      [ 6.  6.  6.  6.  6.  6.]
      [ 7.  7.  7.  7.  7.  7.]
      [ 8.  8.  8.  8.  8.  8.]
      [ 9.  9.  9.  9.  9.  9.]
      [10. 10. 10. 10. 10. 10.]]
    <BLANKLINE>
     [[ 5.  5.  5.  5.  5.  5.]
      [ 6.  6.  6.  6.  6.  6.]
      [ 7.  7.  7.  7.  7.  7.]
      [ 8.  8.  8.  8.  8.  8.]
      [ 9.  9.  9.  9.  9.  9.]
      [10. 10. 10. 10. 10. 10.]]]
    >>> print(coordinates[2])
    [[[1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]]
    <BLANKLINE>
     [[2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]]]
    >>> block_coords, labels = block_split(
    ...     coordinates, block_size=2.5, adjust="region",
    ... )
    >>> print(labels)
    [[[0 0 0 1 1 1]
      [0 0 0 1 1 1]
      [0 0 0 1 1 1]
      [2 2 2 3 3 3]
      [2 2 2 3 3 3]
      [2 2 2 3 3 3]]
    <BLANKLINE>
     [[0 0 0 1 1 1]
      [0 0 0 1 1 1]
      [0 0 0 1 1 1]
      [2 2 2 3 3 3]
      [2 2 2 3 3 3]
      [2 2 2 3 3 3]]]

    """
    coordinates = check_coordinates(coordinates)
    adjust_translation = {"block_size": "spacing", "region": "region"}
    check_adjust(adjust, valid=adjust_translation.keys())
    if region is None:
        region = get_region(coordinates)
    else:
        check_region(region)
    block_coordinates = grid_coordinates(
        region,
        spacing=block_size,
        shape=block_shape,
        adjust=adjust_translation[adjust],
        pixel_register=True,
    )
    tree = KDTree(np.transpose([c.ravel() for c in block_coordinates]))
    labels = tree.query(np.transpose([c.ravel() for c in coordinates]))[1]
    return block_coordinates, labels.reshape(coordinates[0].shape)


def rolling_window(coordinates, window_size, overlap, *, region=None, adjust="overlap"):
    """
    Split points into overlapping windows.

    A window of the given size is moved across the region at a given step
    (specified by the amount of overlap between adjacent windows). Returns the
    indices of points falling inside each window step. You can use the indices
    to select points falling inside a given window.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape.
    window_size : float
        The size of the windows. Units should match the units of *coordinates*.
        In case the window size is not a multiple of the region, either of them
        will be adjusted according to the value of the *adjust* argument.
    overlap : float
        The amount of overlap between adjacent windows. Should be within the
        range 1 > overlap ≥ 0. For example, an overlap of 0.5 means 50%
        overlap. An overlap of 0 will be the same as
        :func:`~bordado.block_split`.
    region : tuple = (W, E, S, N, ...)
        The boundaries of a given region in Cartesian or geographic
        coordinates. If region is not given, will use the bounding region of
        the given coordinates.
    adjust : str = "overlap" or "region"
        Whether to adjust the window overlap or the region, if required.
        Adjusting the overlap or region is required when the combination of
        window size and overlap is not a multiple of the region. Defaults to
        adjusting the overlap.

    Returns
    -------
    window_coordinates : tuple = (easting, northing, ...)
        ND coordinate arrays for the center of each window. Will have the same
        number of arrays as the *coordinates* and each array will have the
        number of dimensions equal to ``len(coordinates)``.
    indices : array
        An array with the same shape as the *window_coordinates*. Each element
        of the array is a tuple of arrays (with the same length as
        *coordinates*) corresponding to the indices of the points that fall
        inside that particular window. Use these indices to index the given
        *coordinates* and select points from a window.

    Examples
    --------
    Generate a set of sample coordinates on a grid to make it easier to
    visualize the windows:

    >>> import bordado as bd
    >>> coordinates = bd.grid_coordinates((-5, -1, 6, 10), spacing=1)
    >>> print(coordinates[0])
    [[-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]]
    >>> print(coordinates[1])
    [[ 6.  6.  6.  6.  6.]
     [ 7.  7.  7.  7.  7.]
     [ 8.  8.  8.  8.  8.]
     [ 9.  9.  9.  9.  9.]
     [10. 10. 10. 10. 10.]]

    Get the coordinates of the centers of rolling windows with 75% overlap and
    an indexer that allows us to select points from each window:

    >>> window_coords, indices = rolling_window(
    ...     coordinates, window_size=2, overlap=0.75,
    ... )

    Window coordinates will be 2D arrays. Their shape is the number of windows
    in each dimension:

    >>> print(window_coords[0].shape, window_coords[1].shape)
    (5, 5) (5, 5)

    The values of these arrays are the coordinates for the center of each
    rolling window:

    >>> print(window_coords[0])
    [[-4.  -3.5 -3.  -2.5 -2. ]
     [-4.  -3.5 -3.  -2.5 -2. ]
     [-4.  -3.5 -3.  -2.5 -2. ]
     [-4.  -3.5 -3.  -2.5 -2. ]
     [-4.  -3.5 -3.  -2.5 -2. ]]
    >>> print(window_coords[1])
    [[7.  7.  7.  7.  7. ]
     [7.5 7.5 7.5 7.5 7.5]
     [8.  8.  8.  8.  8. ]
     [8.5 8.5 8.5 8.5 8.5]
     [9.  9.  9.  9.  9. ]]

    The indices of points falling on each window will have the same shape as
    the window center coordinates:

    >>> print(indices.shape)
    (5, 5)

    Each element of the indices array is a tuple of arrays, one for each
    element in the ``coordinates``:

    >>> print(len(indices[0, 0]))
    2

    They are indices of the points that fall inside the selected window. The
    first element indexes the axis 0 of the coordinate arrays and so forth:

    >>> print(indices[0, 0][0])
    [0 0 0 1 1 1 2 2 2]
    >>> print(indices[0, 0][1])
    [0 1 2 0 1 2 0 1 2]
    >>> print(indices[0, 1][0])
    [0 0 1 1 2 2]
    >>> print(indices[0, 1][1])
    [1 2 1 2 1 2]
    >>> print(indices[0, 2][0])
    [0 0 0 1 1 1 2 2 2]
    >>> print(indices[0, 2][1])
    [1 2 3 1 2 3 1 2 3]

    Use these indices to select the coordinates the points that fall inside
    a window:

    >>> points_window_00 = [c[indices[0, 0]] for c in coordinates]
    >>> print(points_window_00[0])
    [-5. -4. -3. -5. -4. -3. -5. -4. -3.]
    >>> print(points_window_00[1])
    [6. 6. 6. 7. 7. 7. 8. 8. 8.]
    >>> points_window_01 = [c[indices[0, 1]] for c in coordinates]
    >>> print(points_window_01[0])
    [-4. -3. -4. -3. -4. -3.]
    >>> print(points_window_01[1])
    [6. 6. 7. 7. 8. 8.]

    If the coordinates are 1D, the indices will also be 1D:

    >>> coordinates1d = [c.ravel() for c in coordinates]
    >>> window_coords, indices = rolling_window(
    ...     coordinates1d, window_size=2, overlap=0.75,
    ... )
    >>> print(len(indices[0, 0]))
    1
    >>> print(indices[0, 0][0])
    [ 0  1  2  5  6  7 10 11 12]
    >>> print(indices[0, 1][0])
    [ 1  2  6  7 11 12]

    The returned indices can be used in the same way as before to get the same
    coordinates:

    >>> print(coordinates1d[0][indices[0, 0]])
    [-5. -4. -3. -5. -4. -3. -5. -4. -3.]
    >>> print(coordinates1d[1][indices[0, 0]])
    [6. 6. 6. 7. 7. 7. 8. 8. 8.]

    By default, the windows will span the entire data region. You can also
    control the specific region you'd like the windows to cover:

    >>> coordinates = grid_coordinates((-10, 5, 0, 20), spacing=1)
    >>> window_coords, indices = rolling_window(
    ...     coordinates, window_size=2, overlap=0.75, region=(-5, -1, 6, 10),
    ... )

    Even though the data region is larger, our rolling windows should still be
    the same as before:

    >>> print(coordinates[0][indices[0, 1]])
    [-4. -3. -4. -3. -4. -3.]
    >>> print(coordinates[1][indices[0, 1]])
    [6. 6. 7. 7. 8. 8.]

    """
    coordinates = check_coordinates(coordinates)
    adjust_translation = {"overlap": "spacing", "region": "region"}
    check_adjust(adjust, valid=adjust_translation.keys())
    check_overlap(overlap)
    if region is None:
        region = get_region(coordinates)
    else:
        check_region(region)
    # Check if window size is bigger than the minimum dimension of the region
    if window_size > min(region[1] - region[0], region[3] - region[2]):
        message = (
            f"Invalid window size '{window_size}'. Cannot be larger than dimensions of "
            f"the region '{region}'."
        )
        raise ValueError(message)
    # Calculate the region spanning the centers of the rolling windows
    window_region = pad_region(region, -window_size / 2)
    # Calculate the window step based on the amount of overlap
    window_step = (1 - overlap) * window_size
    # Get the coordinates of the centers of each window
    centers = grid_coordinates(
        window_region, spacing=window_step, adjust=adjust_translation[adjust]
    )
    # Use a KD-tree to get the neighbords that fall within half a window size
    # of the window centers.
    tree = KDTree(np.transpose([c.ravel() for c in coordinates]))
    # Coordinates must be transposed because the kd-tree wants them as columns
    # of a matrix. Use p=inf (infinity norm) to get square windows instead of
    # circular ones.
    indices1d = tree.query_ball_point(
        np.transpose([c.ravel() for c in centers]), r=window_size / 2, p=np.inf
    )
    # Make the indices array the same shape as the center coordinates array.
    # That preserves the information of the number of windows in each
    # dimension. Need to first create an empty array of object type because
    # otherwise numpy tries to use the index tuples as dimensions (even if
    # given ndim=1 explicitly). Can't make it 1D and then reshape because the
    # reshape is ignored for some reason. The workaround is to create the array
    # with the correct shape and assign the values to a raveled view of the
    # array.
    indices = np.empty(centers[0].shape, dtype="object")
    # Need to convert the indices to int arrays because unravel_index doesn't
    # like empty lists but can handle empty integer arrays in case a window has
    # no points inside it.
    indices.ravel()[:] = [
        np.unravel_index(np.array(i, dtype="int"), shape=coordinates[0].shape)
        for i in indices1d
    ]
    return centers, indices


def rolling_window_spherical(coordinates, window_size, overlap, *, region=None):
    """
    Split points into overlapping equal area windows on the sphere.

    A window of the given latitudinal size is moved across the region at
    a given step (specified by the amount of overlap between adjacent windows).
    Windows are not regularly distributed on the sphere and are not "square" in
    longitude, latitude. They are evenly spaced in latitude but their
    longitudinal dimension varies to compensate for the convergence of
    meridians at the polar regions. The overlap will also wrap around the 360-0
    longitude discontinuity (see examples below).

    Returns the indices of points falling inside each window step. You can use
    the indices to select points falling inside a given window.

    Parameters
    ----------
    coordinates : tuple = (longitude, latitude)
        Tuple of arrays with the longitude and latitude coordinates of each
        point. Arrays can be Python lists or any numpy-compatible array type.
        Arrays can be of any shape but must all have the same shape.
    window_size : float
        The size of the windows along latitude in decimal degrees. The
        longitudinal window size is adjusted to retain equal area between
        windows. Must be > 0.
    overlap : float
        The amount of overlap between adjacent windows. Should be within the
        range 1 > overlap ≥ 0. For example, an overlap of 0.5 means 50%
        overlap. The overlap may have to be adjusted to make sure windows fit
        inside the given region exactly.
    region : tuple = (W, E, S, N)
        The boundaries of a given region in geographic coordinates. Should have
        a lower and an upper boundary for each dimension of the coordinate
        system. If region is not given, will use the bounding region of the
        given coordinates.

    Returns
    -------
    window_coordinates : tuple = (longitude, latitude)
        1D coordinate arrays for the center of each window.
    indices : array
        1D array with each element of the array being a tuple of 2 arrays
        corresponding to the indices of the points that fall inside that
        particular window. Use these indices to index the given *coordinates*
        and select points from a window.

    Notes
    -----
    Uses the method of [Malkin2016]_ to divide the region into overlapping
    windows of equal area. The windows will have the specified window size in
    latitude but their longitudinal dimensions will be adjusted to account for
    the convergence of meridians at the poles.

    Examples
    --------
    Generate a set of sample coordinates on a grid to make it easier to
    visualize the windows:

    >>> import bordado as bd
    >>> import numpy as np

    >>> coordinates = bd.grid_coordinates((0, 40, 60, 90), spacing=5)
    >>> print(coordinates[0])
    [[ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]
     [ 0.  5. 10. 15. 20. 25. 30. 35. 40.]]
    >>> print(coordinates[1])
    [[60. 60. 60. 60. 60. 60. 60. 60. 60.]
     [65. 65. 65. 65. 65. 65. 65. 65. 65.]
     [70. 70. 70. 70. 70. 70. 70. 70. 70.]
     [75. 75. 75. 75. 75. 75. 75. 75. 75.]
     [80. 80. 80. 80. 80. 80. 80. 80. 80.]
     [85. 85. 85. 85. 85. 85. 85. 85. 85.]
     [90. 90. 90. 90. 90. 90. 90. 90. 90.]]

    Get the coordinates of the centers of rolling windows with 50% overlap and
    an indexer that allows us to select points from each window:

    >>> window_size = 10  # degrees
    >>> window_coords, indices = rolling_window_spherical(
    ...     coordinates, window_size=window_size, overlap=0.5,
    ... )

    Window coordinates will be 1D arrays since the windows aren't regular.
    Their longitudinal size is calculated to preserve their area. Their shape
    is the number of windows generated:

    >>> print(window_coords[0].shape, window_coords[1].shape)
    (8,) (8,)

    The values of these arrays are the coordinates for the center of each
    rolling window. The latitude coordinates will be all at regular intervals
    dictated by the window size and the overlap:

    >>> print(window_coords[1])
    [65. 65. 70. 70. 75. 75. 80. 85.]

    But in longitude, the window sizes (and thus their centers) will spread out
    as latitude increases to balance the convergence of meridians. The window
    size in longitude will be:

    >>> window_size_lon = window_size / np.cos(np.radians(window_coords[1]))
    >>> print(np.array_str(window_size_lon, precision=1))
    [ 23.7  23.7  29.2  29.2  38.6  38.6  57.6 114.7]

    Notice that as we get closer to the pole the window size is larger than the
    region, so in these cases the windows cannot be guaranteed to have equal
    area. The center of each window will be:

    >>> print(np.array_str(window_coords[0], precision=1))
    [11.8 28.2 14.6 25.4 19.3 20.7 20.  20. ]

    If you look closely, you'll see that the amount of overlap between the
    windows isn't exactly 50%. This is because the overlap and window size do
    not result in multiples of the region. So we have to adjust the overlap to
    make things fit. This effect is less evident for smaller windows.

    The indices of points falling on each window will have the same shape as
    the window center coordinates:

    >>> print(indices.shape)
    (8,)

    Each element of the indices array is a tuple of arrays, one for each
    element in the ``coordinates``:

    >>> print(len(indices[0]))
    2

    They are indices of the points that fall inside the selected window (window
    0). The first element indexes the axis 0 of the coordinate arrays and so
    forth. So this:

    >>> print(indices[0][0])
    [0 0 0 0 0 1 1 1 1 1 2 2 2 2 2]

    corresponds to the rows of the coordinate arrays that belong to the first
    window, and this

    >>> print(indices[0][1])
    [0 1 2 3 4 0 1 2 3 4 0 1 2 3 4]

    corresponds to the columns of the coordinate arrays that belong to the
    first window.

    The same can be showed for the second window:

    >>> print(indices[1][0])
    [0 0 0 0 0 1 1 1 1 1 2 2 2 2 2]
    >>> print(indices[1][1])
    [4 5 6 7 8 4 5 6 7 8 4 5 6 7 8]

    and the third window:

    >>> print(indices[-1][0])
    [4 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6 6]
    >>> print(indices[-1][1])
    [0 1 2 3 4 5 6 7 8 0 1 2 3 4 5 6 7 8 0 1 2 3 4 5 6 7 8]

    Use these indices to select the coordinates the points that fall inside
    a window:

    >>> points_window_0 = [c[indices[0]] for c in coordinates]
    >>> print(points_window_0[0])
    [ 0.  5. 10. 15. 20.  0.  5. 10. 15. 20.  0.  5. 10. 15. 20.]
    >>> print(points_window_0[1])
    [60. 60. 60. 60. 60. 65. 65. 65. 65. 65. 70. 70. 70. 70. 70.]

    >>> points_window_1 = [c[indices[1]] for c in coordinates]
    >>> print(points_window_1[0])
    [20. 25. 30. 35. 40. 20. 25. 30. 35. 40. 20. 25. 30. 35. 40.]
    >>> print(points_window_1[1])
    [60. 60. 60. 60. 60. 65. 65. 65. 65. 65. 70. 70. 70. 70. 70.]

    If the coordinates are 1D, the indices will also be 1D:

    >>> coordinates1d = [c.ravel() for c in coordinates]
    >>> window_coords, indices = rolling_window_spherical(
    ...     coordinates1d, window_size=window_size, overlap=0.5,
    ... )
    >>> print(len(indices[0]))
    1

    In this case, the indices will refer to the raveled coordinate array. The
    indexer for the first window will be:

    >>> print(indices[0][0])
    [ 0  1  2  3  4  9 10 11 12 13 18 19 20 21 22]

    And for the second window:

    >>> print(indices[1][0])
    [ 4  5  6  7  8 13 14 15 16 17 22 23 24 25 26]

    The returned indices can be used in the same way as before to get the same
    coordinates:

    >>> print(coordinates1d[0][indices[0]])
    [ 0.  5. 10. 15. 20.  0.  5. 10. 15. 20.  0.  5. 10. 15. 20.]
    >>> print(coordinates1d[1][indices[0]])
    [60. 60. 60. 60. 60. 65. 65. 65. 65. 65. 70. 70. 70. 70. 70.]

    By default, the windows will span the entire data region. You can also
    control the specific region you'd like the windows to cover:

    >>> window_coords, indices = rolling_window_spherical(
    ...     coordinates,
    ...     window_size=window_size,
    ...     overlap=0.5,
    ...     region=(0, 20, 60, 75),
    ... )

    The windows will now try to fit the smaller region instead of the full
    extent of the coordinates:

    >>> print(coordinates[0][indices[0]])
    [ 0.  5. 10. 15. 20.  0.  5. 10. 15. 20.  0.  5. 10. 15. 20.]
    >>> print(coordinates[1][indices[0]])
    [60. 60. 60. 60. 60. 65. 65. 65. 65. 65. 70. 70. 70. 70. 70.]

    >>> print(coordinates[0][indices[1]])
    [ 0.  5. 10. 15. 20.  0.  5. 10. 15. 20.  0.  5. 10. 15. 20.]
    >>> print(coordinates[1][indices[1]])
    [65. 65. 65. 65. 65. 70. 70. 70. 70. 70. 75. 75. 75. 75. 75.]

    If the longitude range is a full 360 degrees, the windows will wrap around
    the 360-0 discontinuity:

    >>> coordinates = bd.grid_coordinates((0, 360, 0, 0), spacing=30)
    >>> print(coordinates[0])
    [[  0.  30.  60.  90. 120. 150. 180. 210. 240. 270. 300. 330. 360.]]
    >>> print(coordinates[1])
    [[0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]]

    >>> window_coords, indices = rolling_window_spherical(
    ...     coordinates, window_size=90, overlap=0.5,
    ... )
    >>> print(indices.shape)
    (8,)
    >>> print(coordinates[0][indices[0]])
    [  0.  30.  60.  90. 360.]

    The 360 point is there because it's the same as the 0 point and so they are
    both inside the window.

    The last window will be centered at 0 and will wrap around the 360-0
    divide:

    >>> print(coordinates[0][indices[-1]])
    [  0.  30. 330. 360.]

    This way, the windows wrap around the globe and overlap all the way around.
    """
    coordinates = check_coordinates(coordinates)
    check_coordinates_geographic(coordinates)
    check_overlap(overlap)
    if region is None:
        region = get_region(coordinates)
    check_region_geographic(region)
    region, coordinates = longitude_continuity(region, coordinates=coordinates)
    if window_size <= 0:
        message = f"Invalid window size '{window_size}'. Must be > 0."
        raise ValueError(message)
    # Check if longitude goes all the way around
    longitude_360 = np.allclose(abs(region[1] - region[0]), 360)
    # Calculate the window step based on the amount of overlap
    window_step = (1 - overlap) * window_size
    # We'll have to do this by bands of latitude. Each band will have
    # a different window size in longitude. See comments below for more
    # details.
    if window_size >= (region[3] - region[2]):
        # If the window size is larger than the region, line_coordinates
        # will through an error.
        bands = np.array([(region[3] + region[2]) / 2])
    else:
        # Always adjust the spacing to avoid falling out of valid geographic
        # boundaries.
        bands = line_coordinates(
            region[2] + window_size / 2,
            region[3] - window_size / 2,
            spacing=window_step,
            adjust="spacing",
        )
    # These will gather the window centers and indices of points inside each
    # window for each band. Window centers will later be concatenated into
    # a single array.
    longitude, latitude, indices1d = [], [], []
    # Need these so we can map the in-band data indices to the overall data
    # indices.
    data_indices = np.arange(coordinates[0].size)
    for central_latitude in bands:
        # Figure out the size in longitude that leads to equal area. See Malkin
        # (2016).
        window_size_lon = window_size / np.cos(np.radians(central_latitude))
        # Calculate the step size for longitude
        window_step_lon = (1 - overlap) * window_size_lon
        # Generate the longitudes of window centers in this band.
        if window_size_lon >= (region[1] - region[0]):
            # If the window size is larger than the region, line_coordinates
            # will through an error.
            band_longitude = np.array([(region[1] + region[0]) / 2])
        else:
            # Always adjust the spacing to make sure windows are evenly
            # distributed across the 360-0 boundary and the region doesn't
            # exceed valid intervals.
            band_longitude = line_coordinates(
                region[0] + window_size_lon / 2,
                region[1] - window_size_lon / 2,
                spacing=window_step_lon,
                adjust="spacing",
            )
        band_latitude = np.full_like(band_longitude, central_latitude)
        # Make a KD tree with points only in this band. This is needed because
        # query_ball_point only works for "square" windows and we'd select too
        # many points if using window_size_lon (with points out of the latitude
        # band) or too little if using window_size (since longitude intervals
        # are larger).
        latitude_min = central_latitude - window_size / 2
        latitude_max = central_latitude + window_size / 2
        in_band = np.logical_and(
            coordinates[1].ravel() >= latitude_min,
            coordinates[1].ravel() <= latitude_max,
        )
        band_coordinates = [c.ravel()[in_band] for c in coordinates]
        tree = KDTree(np.transpose(band_coordinates))
        # Use p=inf (infinity norm) to get square windows instead of circular.
        in_band_indices = tree.query_ball_point(
            np.transpose([band_longitude, band_latitude]),
            r=window_size_lon / 2,
            p=np.inf,
        )
        # Translate the in-band indices to the overall data indices
        band_indices1d = [data_indices[in_band][i] for i in in_band_indices]
        # If the longitude goes around the globe, add an extra window to have
        # continuity around the 360-0 boundary.
        if longitude_360:
            true_window_step_lon = band_longitude[1] - band_longitude[0]
            first_window_longitude = band_longitude[0] - true_window_step_lon
            first_window = (
                first_window_longitude - window_size_lon / 2,
                first_window_longitude + window_size_lon / 2,
                latitude_min,
                latitude_max,
            )
            first_window, band_coordinates = longitude_continuity(
                first_window,
                coordinates=band_coordinates,
            )
            tree = KDTree(np.transpose(band_coordinates))
            in_band_indices = tree.query_ball_point(
                np.array([[first_window_longitude, central_latitude]]),
                r=window_size_lon / 2,
                p=np.inf,
            )
            band_indices1d.extend(data_indices[in_band][i] for i in in_band_indices)
            band_longitude = np.append(band_longitude, first_window_longitude)
            band_latitude = np.append(band_latitude, central_latitude)
        latitude.append(band_latitude)
        longitude.append(band_longitude)
        indices1d.extend(band_indices1d)
    # Join results from all bands
    latitude = np.concatenate(latitude)
    longitude = np.concatenate(longitude)
    centers = (longitude, latitude)
    # Make an indices array for compatibility with the regular rolling window
    # function. But here the window coordinates cannot be a 2D array since
    # longitudinal windows are not regular.
    indices = np.fromiter(
        [
            np.unravel_index(np.array(i, dtype="int"), shape=coordinates[0].shape)
            for i in indices1d
        ],
        dtype="object",
    )
    return centers, indices


def expanding_window(coordinates, center, sizes):
    """
    Select points on windows of expanding size around a center point.

    Produces arrays for indexing the given coordinates to obtain points falling
    inside each window (see examples below). The windows do not necessarily
    have to be expanding in size (the sizes can be in any order).

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Arrays can be
        Python lists or any numpy-compatible array type. Arrays can be of any
        shape but must all have the same shape.
    center : tuple = (easting, northing, ...)
        The coordinates of the center of the window. Must have the same number
        of elements as *coordinates*. Coordinates **cannot be arrays**.
    sizes : array
        The sizes of the windows. Does not have to be in any particular order.
        The order of indices returned will match the order of window sizes
        given. Units should match the units of *coordinates* and *center*.

    Returns
    -------
    indices : list
        Each element of the list corresponds to the indices of points falling
        inside a window. Use them to index the coordinates for each window. The
        indices will depend on the number of dimensions in the input
        coordinates. For example, if the coordinates are 2D arrays, each window
        will contain indices for 2 dimensions (row, column).

    Examples
    --------
    Generate a set of sample coordinates on a grid to make it easier to
    visualize:

    >>> import bordado as bd
    >>> coordinates = bd.grid_coordinates((-5, -1, 6, 10), spacing=1)
    >>> print(coordinates[0])
    [[-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]
     [-5. -4. -3. -2. -1.]]
    >>> print(coordinates[1])
    [[ 6.  6.  6.  6.  6.]
     [ 7.  7.  7.  7.  7.]
     [ 8.  8.  8.  8.  8.]
     [ 9.  9.  9.  9.  9.]
     [10. 10. 10. 10. 10.]]

    Get the expanding window indices (there should be one index per window):

    >>> indices = expanding_window(coordinates, center=(-3, 8), sizes=[1, 2])
    >>> print(len(indices))
    2

    Each element of the indices is a tuple with the arrays that index the
    coordinates that fall inside each window. For example, this is the index of
    the first window (with size 1):

    >>> print(len(indices[0]))
    2
    >>> print(indices[0][0], indices[0][1])
    [2] [2]

    The index have 2 values because the coordinate arrays are 2D, so we need an
    index of the rows and of the columns. We can use them to select points from
    the coordinates that fall inside the first window:

    >>> print(coordinates[0][indices[0]], coordinates[1][indices[0]])
    [-3.] [8.]

    For the other windows, it works the same:

    >>> for index in indices[1]:
    ...     print(index)
    [1 1 1 2 2 2 3 3 3]
    [1 2 3 1 2 3 1 2 3]
    >>> print(coordinates[0][indices[1]])
    [-4. -3. -2. -4. -3. -2. -4. -3. -2.]
    >>> print(coordinates[1][indices[1]])
    [7. 7. 7. 8. 8. 8. 9. 9. 9.]

    Let's make some 1D coordinates to show how this works in that case:

    >>> coordinates1d = tuple(c.ravel() for c in coordinates)

    Getting the indices is the same and there will still be 1 entry per window:

    >>> indices = expanding_window(coordinates1d, center=(-3, 8), sizes=[1, 2])
    >>> print(len(indices))
    2

    But since coordinates are 1D, there is only one index per window (it's
    still in a tuple, though):

    >>> print(len(indices[0]))
    1
    >>> print(indices[0][0])
    [12]

    >>> print(indices[1][0])
    [ 6  7  8 11 12 13 16 17 18]

    The returned indices can be used in the same way as before:

    >>> print(coordinates1d[0][indices[0]], coordinates1d[1][indices[0]])
    [-3.] [8.]

    Coordinates can be more than 2-dimensional as well:

    >>> coordinates3d = bd.grid_coordinates((-5, 0, 5, 10, 1, 2), spacing=1)
    >>> print(coordinates3d[0].shape)
    (2, 6, 6)
    >>> print(coordinates3d[0])
    [[[-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]]
    <BLANKLINE>
     [[-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]
      [-5. -4. -3. -2. -1.  0.]]]
    >>> print(coordinates3d[1])
    [[[ 5.  5.  5.  5.  5.  5.]
      [ 6.  6.  6.  6.  6.  6.]
      [ 7.  7.  7.  7.  7.  7.]
      [ 8.  8.  8.  8.  8.  8.]
      [ 9.  9.  9.  9.  9.  9.]
      [10. 10. 10. 10. 10. 10.]]
    <BLANKLINE>
     [[ 5.  5.  5.  5.  5.  5.]
      [ 6.  6.  6.  6.  6.  6.]
      [ 7.  7.  7.  7.  7.  7.]
      [ 8.  8.  8.  8.  8.  8.]
      [ 9.  9.  9.  9.  9.  9.]
      [10. 10. 10. 10. 10. 10.]]]
    >>> print(coordinates3d[2])
    [[[1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]
      [1. 1. 1. 1. 1. 1.]]
    <BLANKLINE>
     [[2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]
      [2. 2. 2. 2. 2. 2.]]]

    The only difference is that the center coordinates also need to be in
    3-dimensional space (the size of the windows is uniform in all directions):

    >>> indices = expanding_window(
    ...     coordinates3d, center=(-2.5, 8.5, 1.5), sizes=[1, 2],
    ... )
    >>> print(len(indices))
    2

    Each index will have 3 elements, one for each dimension:

    >>> print(len(indices[0]))
    3
    >>> print(indices[0][0])
    [0 0 0 0 1 1 1 1]
    >>> print(indices[0][1])
    [3 3 4 4 3 3 4 4]
    >>> print(indices[0][2])
    [2 3 2 3 2 3 2 3]

    And extracting coordinates for each window also works the same:

    >>> print(coordinates3d[0][indices[0]])
    [-3. -2. -3. -2. -3. -2. -3. -2.]
    >>> print(coordinates3d[1][indices[0]])
    [8. 8. 9. 9. 8. 8. 9. 9.]
    >>> print(coordinates3d[2][indices[0]])
    [1. 1. 1. 1. 2. 2. 2. 2.]

    """
    coordinates = check_coordinates(coordinates)
    shape = coordinates[0].shape
    center = np.atleast_2d(center)
    tree = KDTree(np.transpose([c.ravel() for c in coordinates]))
    indices = []
    for size in sizes:
        # Use p=inf (infinity norm) to get square windows instead of circular
        index1d = tree.query_ball_point(center, r=size / 2, p=np.inf)[0]
        # Convert indices to an array to avoid errors when the index is empty
        # (no points in the window). unravel_index doesn't like empty lists.
        indices.append(np.unravel_index(np.array(index1d, dtype="int"), shape=shape))
    return indices



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions to generate points along a segment between two points.
"""

import numpy as np

from ._line import line_coordinates


def profile_coordinates(
    beginning, end, *, size=None, spacing=None, non_dimensional_coords=None
):
    """
    Generate evenly spaced coordinates along a straight line between points.

    The generated coordinates specify points along a straight line in Cartesian
    space. The points are evenly spaced and can be specified by *size* (number
    of points) or their *spacing*. The points can be n-dimensional.

    Use this function to generates coordinates for sampling along a profile.

    Parameters
    ----------
    beginning : tuple = (easting, northing, ...)
        The coordinates of the starting point of the profile. Coordinates must
        be single values and not array-like.
    end : tuple = (easting, northing, ...)
        The coordinates of the ending point of the profile. Coordinates must be
        single values and not array-like.
    size : int or None
        The number of points in the profile. If None, *spacing* must be
        provided.
    spacing : float or None
        The step size (interval) between points in the profile. If None, *size*
        must be provided.
    non_dimensional_coords : None, scalar, or tuple of scalars
        If not None, then value(s) of extra non-dimensional coordinates
        (coordinates that aren't part of the profile dimensions, like height
        for a lat/lon profile). Will generate extra coordinate arrays from
        these values with the same shape of the final profile coordinates and
        the constant value given here. Use this to generate arrays of constant
        heights or times, for example, which might be needed to accompany
        a profile.

    Returns
    -------
    coordinates : tuple of arrays
        Arrays with coordinates of each point in the profile. Each array
        contains values for a dimension in the order of the given beginning and
        end points, and any extra values given in *non_dimensional_coords*. All
        arrays will be 1-dimensional and have the same shape.
    distances : array
        The straight-line distances between each point in the profile and the
        beginning point.

    Examples
    --------
    Generate a profile between two points with 11 points in it:

    >>> (east, north), dist = profile_coordinates((1, 10), (1, 20), size=11)
    >>> print('easting:', ', '.join(f'{i:.1f}' for i in east))
    easting: 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
    >>> print('northing:', ', '.join(f'{i:.1f}' for i in north))
    northing: 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0
    >>> print('distance:', ', '.join(f'{i:.1f}' for i in dist))
    distance: 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0

    We can equally specify the point spacing instead of the number of points:

    >>> (east, north), dist = profile_coordinates((1, 10), (1, 20), spacing=1)
    >>> print('easting:', ', '.join(f'{i:.1f}' for i in east))
    easting: 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
    >>> print('northing:', ', '.join(f'{i:.1f}' for i in north))
    northing: 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0
    >>> print('distance:', ', '.join(f'{i:.1f}' for i in dist))
    distance: 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0

    The points can also be more than 2-dimensional. The number of returned
    coordinates is the same as the number of input coordinates:

    >>> (east, north, up), dist = profile_coordinates(
    ...     (1, 10, 5), (1, 20, 5), spacing=1,
    ... )
    >>> print('easting:', ', '.join(f'{i:.1f}' for i in east))
    easting: 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
    >>> print('northing:', ', '.join(f'{i:.1f}' for i in north))
    northing: 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0, 18.0, 19.0, 20.0
    >>> print('upward:', ', '.join(f'{i:.1f}' for i in up))
    upward: 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0
    >>> print('distance:', ', '.join(f'{i:.1f}' for i in dist))
    distance: 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0

    It can sometimes be useful to generate an additional array of the same size
    as the coordinates but filled with a single value, for example if doing
    a profile on easting and northing but we also need a constant height value
    returned:

    >>> (east, north, height), dist = profile_coordinates(
    ...     (1, 10), (1, 20), size=11, non_dimensional_coords=35)
    >>> print(height)
    [35. 35. 35. 35. 35. 35. 35. 35. 35. 35. 35.]

    You can specify multiple of these non-dimensional coordinates:

    >>> (east, north, height, time), dist = profile_coordinates(
    ...     (1, 10), (1, 20), size=11, non_dimensional_coords=(35, 0.1))
    >>> print(height)
    [35. 35. 35. 35. 35. 35. 35. 35. 35. 35. 35.]
    >>> print(time)
    [0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1 0.1]
    """
    if len(beginning) != len(end):
        message = (
            "Beginning and end points of the profile must have the same number of "
            f"coordinates. Given {len(beginning)} and {len(end)}, respectively."
        )
        raise ValueError(message)
    difference = np.asarray(end) - np.asarray(beginning)
    point_separation = np.sqrt(np.sum(difference**2))
    directional_vetor = difference / point_separation
    distances = line_coordinates(
        0, point_separation, size=size, spacing=spacing, adjust="spacing"
    )
    coordinates = [
        x + distances * direction for x, direction in zip(beginning, directional_vetor)
    ]
    if non_dimensional_coords is not None:
        for value in np.atleast_1d(non_dimensional_coords):
            coordinates.append(np.full_like(coordinates[0], value))
    return tuple(coordinates), distances


def great_circle_coordinates(
    beginning,
    end,
    *,
    size=None,
    spacing=None,
    radius=6_370_994,
    non_dimensional_coords=None,
):
    """
    Generate evenly spaced coordinates along a great circle between points.

    The beginning and end points must be (longitude, latitude) coordinates on
    a sphere. The generated coordinates will be evenly spaced (in physical
    distances, not degrees) and fall along a great circle path between the two
    points. The points can be specified by *size* (number of points) or their
    *spacing* (in physical units, like meters).

    Use this function to generates coordinates for sampling along a profile
    when data are in geographic coordinates.

    Parameters
    ----------
    beginning : tuple = (longitude, latitude)
        The coordinates of the starting point of the profile. Coordinates must
        be single values and not array-like. Units should be decimal degrees.
    end : tuple = (longitude, latitude)
        The coordinates of the ending point of the profile. Coordinates must be
        single values and not array-like. Units should be decimal degrees.
    size : int or None
        The number of points in the profile. If None, *spacing* must be
        provided.
    spacing : float or None
        The step size (interval) between points in the profile. If None, *size*
        must be provided. Units should be compatible with *radius*
        (usually meters).
    radius : float
        The radius of the sphere, usually the mean radius of the Earth or other
        body used to scale the distances along the great circle. Units should
        be compatible with *spacing* (usually meters). Defaults to the mean
        radius of the WGS84 Earth ellipsoid (6,370,994 meters).
    non_dimensional_coords : None, scalar, or tuple of scalars
        If not None, then value(s) of extra non-dimensional coordinates
        (coordinates that aren't part of the profile dimensions, like height
        for a lat/lon profile). Will generate extra coordinate arrays from
        these values with the same shape of the final profile coordinates and
        the constant value given here. Use this to generate arrays of constant
        heights or times, for example, which might be needed to accompany
        a profile.

    Returns
    -------
    coordinates : tuple = (longitude, latitude, ...)
        Arrays with the coordinates of each point in the profile. The first two
        are longitude and latitude. Subsequent arrays are any extra values
        given in *non_dimensional_coords*. All arrays will be 1-dimensional and
        have the same shape.
    distances : array
        The great circle distances between each point in the profile and the
        beginning point.

    Examples
    --------
    Generate coordinates between points at the equator for a sphere with a unit
    radius:

    >>> import numpy as np
    >>> spacing = 2 * np.pi / 180
    >>> (longitude, latitude), distance = great_circle_coordinates(
    ...     (0, 0), (10, 0), spacing=spacing, radius=1,
    ... )
    >>> print('longitude:', ', '.join(f'{i:.1f}' for i in longitude))
    longitude: 0.0, 2.0, 4.0, 6.0, 8.0, 10.0
    >>> print('latitude:', ', '.join(f'{i:.1f}' for i in latitude))
    latitude: 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
    >>> print('distance:', ', '.join(f'{i:.4f}' for i in distance))
    distance: 0.0000, 0.0349, 0.0698, 0.1047, 0.1396, 0.1745
    >>> print(np.allclose(distance[1:] - distance[0:-1], spacing))
    True

    It can sometimes be useful to generate an additional array of the same size
    as the coordinates but filled with a single value, for example if  we also
    need a constant height value returned:

    >>> (lon, lat, height), dist = great_circle_coordinates(
    ...     (1, 10), (1, 20), size=11, non_dimensional_coords=35)
    >>> print(height)
    [35. 35. 35. 35. 35. 35. 35. 35. 35. 35. 35.]
    """
    if len(beginning) != 2 or len(end) != 2:
        message = (
            "Beginning and end points of a great circle profile must have two "
            "coordinates (longitude and latitude). "
            f"Given {beginning} and {end}, respectively."
        )
        raise ValueError(message)
    lon1, lat1 = np.radians(beginning)
    lon2, lat2 = np.radians(end)
    coslon = np.cos(lon2 - lon1)
    sinlon = np.sin(lon2 - lon1)
    coslat1 = np.cos(lat1)
    sinlat1 = np.sin(lat1)
    coslat2 = np.cos(lat2)
    sinlat2 = np.sin(lat2)
    # These are needed to calculate the lon/lat coordinates of the profile.
    # See https://en.wikipedia.org/wiki/Great-circle_navigation#Finding_way-points
    azimuth1 = np.arctan2(
        coslat2 * sinlon, coslat1 * sinlat2 - sinlat1 * coslat2 * coslon
    )
    sinazimuth1 = np.sin(azimuth1)
    cosazimuth1 = np.cos(azimuth1)
    azimuth_equator = np.arctan2(
        sinazimuth1 * coslat1, np.sqrt(cosazimuth1**2 + sinazimuth1**2 * sinlat1**2)
    )
    sinazimuth_equator = np.sin(azimuth_equator)
    cosazimuth_equator = np.cos(azimuth_equator)
    great_circle_equator = np.arctan2(np.tan(lat1), cosazimuth1)
    lon_equator = lon1 - np.arctan2(
        sinazimuth_equator * np.sin(great_circle_equator), np.cos(great_circle_equator)
    )
    # The great-circle distance between start and end (in radians)
    # This is the haversine formula: https://en.wikipedia.org/wiki/Haversine_formula
    great_circle_distance = 2 * np.arcsin(
        np.sqrt(
            np.sin((lat2 - lat1) / 2) ** 2
            + coslat1 * coslat2 * np.sin((lon2 - lon1) / 2) ** 2
        )
    )
    # Generate evenly spaced points along the great circle.
    # Multiply by the radius so the distance is compatible with the spacing but
    # divide after because calculations below are for unit radius.
    distances = (
        line_coordinates(
            0,
            great_circle_distance * radius,
            size=size,
            spacing=spacing,
            adjust="spacing",
        )
        / radius
    )
    # Make the distances relative to where the great circle cross the equator
    # This is needed for the calculations below.
    distances_equator = distances + great_circle_equator
    sindistances_equator = np.sin(distances_equator)
    cosdistances_equator = np.cos(distances_equator)
    # Calculate the lon/lat coordinates of each point given their arc distance
    # and the azimuth of the great circle
    latitudes = np.degrees(
        np.arctan2(
            cosazimuth_equator * sindistances_equator,
            np.sqrt(
                cosdistances_equator**2
                + (sinazimuth_equator * sindistances_equator) ** 2
            ),
        )
    )
    longitudes = np.degrees(
        lon_equator
        + np.arctan2(sinazimuth_equator * sindistances_equator, cosdistances_equator)
    )
    coordinates = [longitudes, latitudes]
    if non_dimensional_coords is not None:
        for value in np.atleast_1d(non_dimensional_coords):
            coordinates.append(np.full_like(coordinates[0], value))
    # Convert the arc-distances into meters
    distances *= radius
    return tuple(coordinates), distances



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Functions to calculate the distances between neighbors.
"""

import numpy as np
import scipy.spatial

from ._validation import check_coordinates


def neighbor_distance_statistics(coordinates, statistic, *, k=1):
    """
    Calculate statistics of the distances to the k-nearest neighbors of points.

    For each point specified in *coordinates*, calculate the given statistic on
    the Cartesian distance to its *k* neighbors among the other points in the
    dataset.

    Useful for finding mean/median distances between points, general point
    spread (standard deviation), variability of neighboring distances
    (peak-to-peak), etc.

    Parameters
    ----------
    coordinates : tuple = (easting, northing, ...)
        Tuple of arrays with the coordinates of each point. Should be in an
        order compatible with the order of boundaries in *region*. Arrays can
        be Python lists. Arrays can be of any shape but must all have the same
        shape.
    statistic : str
        Which statistic to calculate for the distances of the k-nearest
        neighbors of each point. Valid values are: ``"mean"``,  ``"median"``,
        ``"std"`` (standard deviation),  ``"var"`` (variance),  ``"ptp"``
        (peak-to-peak amplitude).
    k : int
        Will calculate the median of the *k* nearest neighbors of each point. A
        value of 1 will result in the distance to nearest neighbor of each data
        point. Must be >= 1. Default is 1.

    Returns
    -------
    statistics : array
        An array with the statistic of the k-nearest neighbor distances of each
        point. The array will have the same shape as the input coordinate
        arrays.

    Raises
    ------
    ValueError
        If *k* is less than 1, if the *statistic* is invalid, or if coordinate
        arrays have different shapes.

    Notes
    -----
    To get the average point spacing for sparse uniformly spaced datasets,
    calculating the mean/median using *k* of 1 is reasonable. Datasets with
    points clustered into tight groups (e.g., densely sampled along a flight
    line or ship track) will have very small distances to the closest
    neighbors, which is not representative of the actual median spacing of
    points because it doesn't take the spacing between lines into account. In
    these cases, a median of the 10-20 or more nearest neighbors might be more
    representative.

    Examples
    --------
    Generate a grid of points for an example:

    >>> import bordado as bd
    >>> coordinates = bd.grid_coordinates((5, 10, -20, -17), spacing=1)
    >>> print(coordinates[0])
    [[ 5.  6.  7.  8.  9. 10.]
     [ 5.  6.  7.  8.  9. 10.]
     [ 5.  6.  7.  8.  9. 10.]
     [ 5.  6.  7.  8.  9. 10.]]
    >>> print(coordinates[1])
    [[-20. -20. -20. -20. -20. -20.]
     [-19. -19. -19. -19. -19. -19.]
     [-18. -18. -18. -18. -18. -18.]
     [-17. -17. -17. -17. -17. -17.]]

    The mean of the distance to 1 nearest neighbor should be the grid spacing:

    >>> mean_distances = neighbor_distance_statistics(coordinates, "mean", k=1)
    >>> print(mean_distances)
    [[1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]]

    The statistics returned have the same shape as the input coordinates:

    >>> print(mean_distances.shape, coordinates[0].shape)
    (4, 6) (4, 6)
    >>> mean_distances = neighbor_distance_statistics(
    ...     [c.ravel() for c in coordinates], "mean", k=1,
    ... )
    >>> print(mean_distances.shape)
    (24,)

    The mean distance to the 2 nearest points should also all be 1 since they
    are the neighbors along the rows and columns of the matrix:

    >>> mean_distances = neighbor_distance_statistics(coordinates, "mean", k=2)
    >>> print(mean_distances)
    [[1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]]

    The distance to the 3 nearest points is 1 but on the corners of the grid,
    the distances are [1, 1, sqrt(2)] which leads to a median of 1:

    >>> median_distances = neighbor_distance_statistics(
    ...     coordinates, "median", k=3,
    ... )
    >>> print(median_distances)
    [[1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]
     [1. 1. 1. 1. 1. 1.]]

    But using the 4 nearest points leads to distances [1, 1, sqrt(2), 2] at the
    corners, which results in a median of 1.21.

    >>> median_distances = neighbor_distance_statistics(
    ...     coordinates, "median", k=4,
    ... )
    >>> for line in median_distances:
    ...     print(" ".join([f"{i:.2f}" for i in line]))
    1.21 1.00 1.00 1.00 1.00 1.21
    1.00 1.00 1.00 1.00 1.00 1.00
    1.00 1.00 1.00 1.00 1.00 1.00
    1.21 1.00 1.00 1.00 1.00 1.21

    """
    coordinates = check_coordinates(coordinates)
    if k < 1:
        message = f"Invalid number of neighbors 'k={k}'. Must be >= 1."
        raise ValueError(message)
    statistics = {
        "mean": np.mean,
        "median": np.median,
        "std": np.std,
        "var": np.var,
        "ptp": np.ptp,
    }
    if statistic not in statistics:
        message = (
            f"Invalid statistic '{statistic}'. Must be one of: {statistics.keys()}."
        )
        raise ValueError(message)
    shape = np.broadcast(*coordinates).shape
    transposed_coordinates = np.transpose([c.ravel() for c in coordinates])
    tree = scipy.spatial.KDTree(transposed_coordinates)
    # The k=1 nearest point is going to be the point itself (with a distance of
    # zero) because we don't remove each point from the dataset in turn. We
    # don't care about that distance so start with the second closest. Only get
    # the first element returned (the distance) and ignore the rest (the
    # neighbor indices).
    k_distances = tree.query(transposed_coordinates, k=k + 1)[0][:, 1:]
    result = statistics[statistic](k_distances, axis=1)
    return result.reshape(shape)



# Copyright (c) 2025 The Bordado Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
These are the functions and classes that make up the Bordado API.
"""

from ._distance import neighbor_distance_statistics
from ._grid import grid_coordinates
from ._line import line_coordinates
from ._profile import great_circle_coordinates, profile_coordinates
from ._random import random_coordinates, random_coordinates_spherical
from ._region import get_region, inside, pad_region
from ._split import (
    block_split,
    expanding_window,
    rolling_window,
    rolling_window_spherical,
)
from ._utils import shape_to_spacing, spacing_to_size
from ._version import __version__

# Append a leading "v" to the generated version by setuptools_scm
__version__ = f"v{__version__}"
